from datetime import datetime
from typing import Generic, TypeAlias, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from ._base import BaseSchema
from .geo import LocationEleSchema


class SourcePropertiesSchema(BaseModel):
    """Properties saved together with the source data.

    Examples:
        See `hut_services.osm.schema.OsmProperties`."""

    # See [`OsmProperties`][hut_services.osm.schema.OsmProperties]."""


class SourceDataSchema(BaseSchema):
    """SourceData schema"""

    def get_id(self) -> str:
        """Get source `id`."""
        raise NotImplementedError("'get_id()' not implemented")

    def get_name(self) -> str:
        """Get source hut `name`."""
        raise NotImplementedError("'get_name()' not implemented")


TSourceData_co = TypeVar("TSourceData_co", bound=SourceDataSchema, covariant=True)
TProperties_co = TypeVar("TProperties_co", bound=SourcePropertiesSchema, covariant=True)


class BaseHutSourceSchema(BaseModel, Generic[TSourceData_co, TProperties_co]):
    """Base class for a hut source.

    Attributes:
        source_name: Name of the source (e.g. osm, wikipedia, ...).
        name: Original hut name.
        location: Location of the hut.
        source_id: Originial source id of the hut.
        source_data: Source data for this hut.
        source_properties: Additinal source data properties.
        version: Version of the service when this entry was created.
        created: Created.

    Examples:
        ``` py
        class MyHutSource(BaseHutSourceSchema[MyHutSchema, SourcePropertiesSchema]):
            source_name: str = "my"
        ```

        Or with different properties: `hut_services.osm.schema.OsmHutSource`.
    """

    # Or with different properties: [`OsmHutSource`][hut_services.osm.schema.OsmHutSource].

    model_config = ConfigDict(coerce_numbers_to_str=True)

    source_name: str = Field("unknown", description="Name of the source (e.g. osm, wikipedia, ...).")

    # hut information
    name: str = Field(..., description="Original hut name.")
    location: LocationEleSchema | None = Field(None, description="Location of the hut.")

    # source information
    source_id: str = Field(..., description="Originial source id of the hut.")
    source_data: TSourceData_co | None = Field(None, description="Source data for this hut.")
    source_properties: TProperties_co | None = Field(None, description="Additinal source data properties.")

    # create information
    version: int = Field(default=0, description="Version of the service when this entry was created.")
    created: datetime = Field(
        default_factory=datetime.now, description="Version of the service when this entry was created."
    )

    def __str__(self) -> str:
        loc = f"({self.location.lon},{self.location.lat})" if self.location is not None else "(no location)"
        return f"<{self.source_name} #{self.source_id} - {self.name} {loc}>"

    @property
    def source_properties_schema(self) -> dict:
        """Returns JSON schema for the 'source_properties' fields.

        Returns:
            JSON schema."""
        if self.source_properties is not None:
            return self.source_properties.model_json_schema(by_alias=True)
        return {}

    def show(
        self,
        source_id: bool = True,
        location: bool = True,
        elevation: bool = True,
        source_name: bool = True,
        version: bool = False,
        created: bool = False,
    ) -> str:
        """Returns a formatted string with the hut information which can be printed.

        Args:
            source_id: Show source ID.
            location:  Show location.
            elevation: Show elevation.
            source_name: Show source name.
            version: Show verions.
            created: Show created date.

        Returns:
            Formatted string.

        """
        out = [f"{self.name}"]
        if source_id:
            out.append(f"  id:        {self.source_id}")
        if location:
            loc = f"{self.location.lon},{self.location.lat}" if self.location is not None else "not set"
            out.append(f"  location:  {loc}")
        if elevation:
            ele = self.location.ele if self.location is not None else "not set"
            out.append(f"  elevation: {ele}")
        if source_name:
            out.append(f"  source:    {self.source_name}")
        if version:
            out.append(f"  version:   {self.version}")
        if created:
            out.append(f"  created:   {self.created}")
        return "\n".join(out)


HutSourceSchema: TypeAlias = BaseHutSourceSchema[SourceDataSchema, SourcePropertiesSchema]
"""Hut source schema, used for typing.
With pydantic `BaseModel` and [`SourcePrpertiesSchema`][hut_services.core.schema.SourcePropertiesSchema]."""

# class HutSourceSchema(BaseHutSourceSchema[BaseModel, SourcePropertiesSchema]):
#    """Hut source schema, used for typing.
#    With pydantic `BaseModel` and [`SourcePrpertiesSchema`][hut_services.core.schema.SourcePropertiesSchema]."""



from datetime import datetime

from pydantic import Field
from pydantic_string_url import HttpUrl

from ._base import BaseSchema
from ._license import AuthorSchema, LicenseSchema, SourceSchema
from .locale import TranslationSchema


class PhotoSchema(BaseSchema):
    """Schema for a hut photo."""

    licenses: list[LicenseSchema]
    caption: TranslationSchema = Field(default_factory=TranslationSchema)
    source: SourceSchema | None
    author: AuthorSchema | None
    comment: str = Field("", max_length=20000, description="Additional private comment, e.g for review.")
    raw_url: HttpUrl = Field(..., description="Url to the raw image, this can be used to download or embed the image.")
    width: int = Field(..., description="Image width in pixels.")
    height: int = Field(..., description="Image height in pixels.")
    url: HttpUrl = Field(
        ...,
        description="Url to the image on the side, this should not be used to include it direclty, rather to just link to it.",
    )
    capture_date: datetime | None
    tags: set[str] | None = None



import re
import sys
from typing import Literal

from tasks import Ctx, echo, env, error, header, info, success, task, warning  # noqa: F401


@task(
    default=True,
    help={
        "unreleased": "Get the unreleased, not published entries. Overwrites --version",
        "version": "Version to extract, either number or 'current' for the current version",
        "plain": "Do not print headers or warnings",
    },
)
def changelog(
    c: Ctx,
    unreleased: bool = False,
    version: str | Literal["current", "unreleased"] = "current",
    plain: bool = False,
):
    """Get changelog entries for a specific version"""
    if version == "unreleased" or unreleased:
        # get it from current PRs, not in CHANGELOG yet.
        content = c.run("git-cliff --unreleased --bump --strip all | tail -n +2", hide=True).stdout.strip()
        bumped_version = c.run("git cliff --bumped-version", hide=True).stdout.strip().replace("v", "")
        captured_version = f"latest: {bumped_version}"
    else:
        input_file = "CHANGELOG.md"

        try:
            with open(input_file) as file:
                lines = file.readlines()
        except FileNotFoundError:
            error(f"File {input_file} not found.")
        except Exception as e:
            error(f"An error occurred: {e}")

        # Initialize variables to capture the desired section
        capture = False
        extracted_content = []
        version_header_pattern = "## \\[(.*)\\]\\s+.*" if version == "current" else f"## \\[({version})\\]\\s+.*"

        for line in lines:
            match = re.match(version_header_pattern, line)
            if match and not capture:
                captured_version = match.group(1)
                capture = True
                continue

            if capture:
                # Stop capturing if the next version is reached
                if re.match(r"^## \[.*\].*", line):
                    break
                extracted_content.append(line)
        content = "".join(extracted_content).strip().strip("\n")
    if not content:
        warning(f"Nothing found for '{version}'") if not plain else None
        sys.exit(1)
    else:
        header(f"Changelog for '{captured_version}'") if not plain else None
        echo(content, raw=True)
        header() if not plain else None



#!/usr/bin/env python
# from functools import lru_cache
import logging
import re
import typing as t
from datetime import datetime
from typing import Any, cast
from urllib.parse import quote, urlparse

import defusedxml.ElementTree
import requests

# from typing import Any, Literal, Mapping
from bs4 import BeautifulSoup, Tag
from pydantic import ValidationError
from pydantic_string_url import HttpUrl
from rich import print as rprint

from hut_services import (
    AuthorSchema,
    LicenseSchema,
    PhotoSchema,
    SourceSchema,
    TranslationSchema,
    file_cache,
)
from hut_services.core.schema import HutSchema
from hut_services.core.schema.geo import BBox

if __name__ == "__main__":  # only for testing
    from icecream import ic  # type: ignore[import-untyped] # noqa: F401, RUF100 , PGH003
    from rich import print as rprint  # noqa: F401, RUF100
    from rich.traceback import install

    install(show_locals=False)


logger = logging.getLogger(__name__)


def _parse_href_html_field(html: str | None) -> tuple[str, HttpUrl | None]:
    """Parse an HTML field to extract plain text and URL (if present)."""
    if not html:
        return "", None
    soup = BeautifulSoup(html, "html.parser")
    if soup.a:
        name = soup.a.text or ""
        url = soup.a.get("href")
        if isinstance(url, list):
            url = url[0]
        url_cast = t.cast(HttpUrl, url) if url is not None else None
        return name, url_cast
    else:
        return html.strip(), None


def _parse_time_html_field(html: str | None) -> datetime | None:
    """
    Parse an HTML field to extract a datetime object.

    Args:
        html: The HTML field to parse.

    Returns:
        The extracted datetime object, or None if parsing fails.
    """
    if html is None:
        return None
    soup = BeautifulSoup(html, "html.parser")
    time_tag = soup.find("time")
    if time_tag and isinstance(time_tag, Tag):
        datetime_str = str(time_tag.get("datetime"))
        try:
            return datetime.strptime(datetime_str, "%Y-%m-%d")
        except ValueError:
            try:
                return datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
            except ValueError:
                # Handle other possible date formats or raise an error
                return None
    return None


def _extract_hostname(url: str) -> str:
    """Extract the hostname from a full URL."""
    parsed_url = urlparse(url)
    return parsed_url.hostname if parsed_url.hostname else url


def _resize_image_url(url: str, max_dimension: int) -> str:
    """Resize the image URL if width or height exceeds max_dimension."""
    filename = url.split("/")[-1]
    return (url + f"/{max_dimension}px-{filename}").replace("commons/", "commons/thumb/")


@file_cache()
def _wikicommon_api_call(
    filename: str, api_url: str = "https://magnus-toolserver.toolforge.org/commonsapi.php"
) -> bytes | None:
    """Fetch image information from Magnus Toolserver API and return structured data using Pydantic."""
    api_url_image = f"{api_url}?image={filename}"
    response = requests.get(api_url_image, timeout=20)

    if response.status_code != 200:
        err_msg = f"Error fetching data from Magnus Toolserver: {response.status_code}"
        raise Exception(err_msg)  # noqa: TRY002
    return response.content if isinstance(response.content, bytes) else None


def get_wikicommon_photo_info(
    filename: str, api_url: str = "https://magnus-toolserver.toolforge.org/commonsapi.php", max_dimension: int = 3000
) -> PhotoSchema:
    """Fetch image information from Magnus Toolserver API and return structured data using Pydantic."""
    content = _wikicommon_api_call(filename, api_url)

    # Parse XML response
    root = defusedxml.ElementTree.fromstring(content)

    # Extract fields from XML
    file_info = root.find("file")

    def file_info_find(name: str, default: str | None = None, elem: Any = None) -> str | None:
        elem = file_info if elem is None else elem
        if elem is not None:
            return elem.find(name).text if elem.find(name) is not None else default
        return default

    assert file_info is not None  # noqa: S101

    # Image URL and dimensions
    image_url = file_info_find("urls/file")
    title = file_info_find("title")
    width = int(cast(str, file_info_find("width")))
    height = int(cast(str, file_info_find("height")))
    img_date = _parse_time_html_field(cast(str, file_info_find("date")))

    wikicommons_url = f"https://commons.wikimedia.org/wiki/{quote(title or '')}"
    # Resize the image URL if width or height is greater than max_dimension
    if image_url and width and height and (width > max_dimension or height > max_dimension):
        image_url = _resize_image_url(image_url, max_dimension)
        # update height
        orig_height = height
        orig_width = width
        height = max_dimension if orig_height > orig_width else round(orig_height / orig_width * max_dimension)
        width = max_dimension if orig_width > orig_height else round(orig_width / orig_height * max_dimension)
    # License information
    license_elems = root.findall("licenses/license")
    license_name = "No license available"
    license_slug = "missing"
    license_info_url: HttpUrl | None = None
    licenses = []
    for license_elem in license_elems:
        license_name = cast(
            str, file_info_find("name", file_info_find("full_name", "missing", license_elem), license_elem)
        )
        license_name = license_name.replace("migrated", "").strip(" -")
        license_slug = license_name.lower()
        license_info_url = cast(HttpUrl | None, file_info_find("license_info_url", None, license_elem))
        if "cc" in license_slug.lower():
            license_slug = license_slug.split(",")[0].strip()
        licenses.append(LicenseSchema(slug=license_slug, url=license_info_url, name=license_name))

    # Captions only
    captions = TranslationSchema()
    captions_elem = root.find("description")
    if captions_elem is not None:
        for lang_elem in captions_elem.findall("language"):
            lang_code = lang_elem.get("code")
            caption = lang_elem.text.strip() if lang_elem.text else ""
            if lang_code in ["en", "de", "it", "fr"]:
                setattr(captions, lang_code, caption)

    # Author and Source
    author_raw = file_info_find("author", "")
    source_raw = file_info_find("source", wikicommons_url)

    # Parse author and source
    author_name, author_url = _parse_href_html_field(author_raw)
    source_name, source_url = _parse_href_html_field(source_raw)

    # Handle 'int-own-work' source
    source_ident = title
    # if "int-own-work" in source_name:
    #    source_url = cast(HttpUrl, wikicommons_url)
    #    source_name = "wikicommons"
    comment = f"* Wikicommons source: {wikicommons_url}\n"
    if "camptocamp" in source_name.lower():
        source_name = "camptocamp"
        match = re.search(r"\d{3,}", cast(str, source_url))
        source_ident = match.group() if match else source_ident
        match = re.search(r"http:.*/\d{3,}", cast(str, source_url))
        source_url = cast(HttpUrl, match.group()) if match else source_url
    elif "refuges.info" in source_name.lower():
        source_name = "refuges.info"
        match = re.search(r"(\d{3,})-originale", cast(str, source_url))
        if match:
            source_ident = match.groups()[0]
        else:
            match = re.search(r"(#C\d{3,})", cast(str, source_url))
            source_ident = match.groups()[0] if match else source_ident
    else:
        if source_name:
            comment += f"* Original source name: {source_name}\n"
        if source_url:
            comment += f"* Original source url: {source_url}\n"
        source_url = cast(HttpUrl, wikicommons_url)
        source_name = "wikicommons"
    # if not source_url:
    #    source_url = cast(HttpUrl, wikicommons_url)

    author = AuthorSchema(name=author_name, url=author_url) if author_name else None
    source = SourceSchema(url=source_url, name=source_name, ident=source_ident)

    # Return structured data using Pydantic
    return PhotoSchema(
        licenses=licenses,
        caption=captions,
        author=author,
        source=source,
        raw_url=image_url,
        width=width,
        height=height,
        url=wikicommons_url,
        capture_date=img_date,
        comment=comment.strip(),
    )


class WikicommonsService:
    """Service to get photo from
    [Wikimedia Commons](https://commons.wikimedia.org).

    Note:
        This is only used to get photo information, not to get huts!
    """

    def __init__(
        self,
        request_url: str = "https://magnus-toolserver.toolforge.org/commonsapi.php",
        max_dimension: int = 3600,
    ):
        super().__init__()
        self.request_url = request_url
        self._max_dimension = max_dimension

    def get_photo(self, filename: str) -> PhotoSchema:
        return get_wikicommon_photo_info(filename=filename, api_url=self.request_url, max_dimension=self._max_dimension)

    def get_huts_from_source(self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: dict) -> list:
        raise NotImplementedError("Get huts from source not implemented for WikiCommons.")

    def convert(self, src: t.Mapping | t.Any) -> HutSchema:
        raise NotImplementedError("Get huts from source not implemented for WikiCommons.")


wikicommons_service = WikicommonsService()

if __name__ == "__main__":
    logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

    # File names to query
    file_names = ["Wildhornhuette.jpg", "Lohner hut SAC.jpg", "RifugioVallanta.jpg", "Refuge d'Ambin.jpeg"]

    for file_name in file_names:
        try:
            image_info = wikicommons_service.get_photo(file_name)
            rprint(image_info)
        except ValidationError as e:
            rprint("Validation error:", e)
        # except Exception as ex:
        #    print("Error:", ex)



from invoke.collection import Collection
from invoke.context import Context as Ctx
from invoke.tasks import task

from tasks._env import EnvError, env
from tasks._logger import doc, echo, error, header, info, success, warning
from tasks import changelog, check, docs, project, tests


@task
def help(c: Ctx):  # noqa: A001
    """Show this help"""
    c.run("inv --list", pty=True)


ns = Collection(
    project.install,
    project.release,
    project.version,
    project.update_venv,
    docs,
    tests,
    check,
    changelog.changelog,
    help,
)


__all__ = (
    "Ctx",
    "EnvError",
    "doc",
    "echo",
    "env",
    "error",
    "header",
    "info",
    "install",
    "success",
    "warning",
)



from tasks import Ctx, EnvError, echo, env, error, header, info, success, task, warning  # noqa: F401


@task(help={"port": "Port to serve documentation locally, default: '8000'"})
def serve(c: Ctx, port: int = 8000):
    """Serve docu for preview."""
    dev_addr = f"localhost:{port}"
    c.run(f"mkdocs serve --dev-addr {dev_addr} -w src", echo=True, pty=True)


@task
def build(c: Ctx):
    """Build docu."""
    c.run("mkdocs build", echo=True, pty=True)
    success("Documentation build in 'site' directory.")
    info("Run 'python -m http.server -d site' for testing (not in production).")



import datetime
import typing as t

from hut_services import HutSourceSchema, clear_file_cache
from hut_services.core.schema import HutBookingsSchema, HutSchema
from hut_services.core.schema.geo import BBox

THutSourceSchema = t.TypeVar("THutSourceSchema", bound=HutSourceSchema, covariant=True)


class BaseService(t.Generic[THutSourceSchema]):
    """Base service which is inheritated by other services.

    Warning:
        Do not use this directly.

    The following attributes are used to define which paramets are supported
    by a service.

    Attributes:
        support_bbox: Support for `bbox` as parameter.
        support_limit: Support for `limit` as parameter.
        support_offset: Support for `offset` as parameter.
        support_convert: Support for `convert` as parameter.

    Examples:
        Custom service base in `BaseService`.
        For this the two schemas `MyHutSource` and `MyInfoHutConvert` need to be defined as well.
        ```python
        from typing import Any
        from hut_services.core.schema.geo import BBox
        from hut_services import BaseService, HutSchema

        # TODO: define this somewhere:
        from my_service.schema import MyHutSource, MyInfoHutConvert

        class MyService(BaseService[MyHutSource]):
            def __init__(self, request_url: str = "http://hut.info"):
                super().__init__(support_bbox=True,
                                 support_limit=True,
                                 support_offset=True,
                                 support_convert=True)
                self.request_url = request_url

            def get_huts_from_source(
                self, bbox: BBox | None = None, limit: int = 1,
                offset: int = 0, **kwargs: Any
            ) -> list[MyHutSource]:
                src_huts = httpx.get(self.request_url)
                return [MyHutSource(**h) for h in src_huts]

            def convert(self, src: MyHutSource) -> HutSchema:
                return MyInfoHutConvert(source=src.source_data).get_hut()
        ```
    """

    support_bbox: bool = False
    support_limit: bool = False
    support_offset: bool = False
    support_convert: bool = False
    support_booking: bool = False

    def __init__(
        self,
        support_bbox: bool = False,
        support_limit: bool = False,
        support_offset: bool = False,
        support_convert: bool = False,
        support_booking: bool = False,
    ) -> None:
        self.support_bbox = support_bbox
        self.support_limit = support_limit
        self.support_offset = support_offset
        self.support_convert = support_convert
        self.support_booking = support_booking

    @classmethod
    def clear_all_cache(cls) -> None:
        """Clears the cache of all services!"""
        clear_file_cache()

    def get_huts_from_source(
        self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: t.Any
    ) -> list[THutSourceSchema]:
        """Get all huts from source.

        Args:
            bbox: Boundary box.
            limit: Limit (how many entries to retrieve).
            offset: Offset of the request.

        Returns:
            Huts from source.
        """
        raise self.MethodNotImplementedError(self, "get_huts_from_source")

    def convert(self, src: t.Mapping | t.Any, include_photos: bool = True) -> HutSchema:
        """Convert one hut from source to [`HutSchema`][hut_services.HutSchema].

        Args:
            src: Source schema.
            include_photos: Include photos, some service need additonal requests to get the photos.

        Returns:
            Converted hut.
        """
        # example:
        # hut_src = OsmHutSource(**src) if isinstance(src, t.Mapping) else OsmHutSource.model_validate(src)
        raise self.MethodNotImplementedError(self, "convert")

    def get_huts(
        self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, include_photos: bool = True, **kwargs: t.Any
    ) -> list[HutSchema]:
        """Get all huts form source and converts them.
        Calls [`get_huts_from_source()`][hut_services.BaseService.get_huts_from_source]
        and [`convert()`][hut_services.BaseService.convert].

        Returns:
            Converted huts from source."""
        huts = []
        src_huts = self.get_huts_from_source(bbox=bbox, limit=limit, offset=offset, **kwargs)
        huts = [self.convert(h, include_photos=include_photos) for h in src_huts]
        return huts

    def get_bookings(
        self,
        date: datetime.datetime | datetime.date | t.Literal["now"] | None = None,
        days: int | None = None,
        source_ids: list[int] | None = None,
        lang: str = "de",
    ) -> dict[int, HutBookingsSchema]:
        """Get bookings for a list of huts.

        Args:
            date: Start daye for the bookings
            days: Duration in days
            source_ids: A list of ids to return (source id, not the hut id), if set to `None` all are returned

        Returns:
            A dictionary with the bookings (key = soruce id).
        """
        raise self.MethodNotImplementedError(self, "get_bookings")

    class MethodNotImplementedError(NotImplementedError):
        """Method is not implemented exception.

        Args:
            obj: Service object (e.g. `MyService`).
            method: Method which is not implemented.
        """

        def __init__(self, obj: "BaseService", method: str):
            class_name = str(obj.__class__).replace("<class '", "").replace("'>", "")
            message = f"Service '{class_name}' method '{method}' is not implemented."
            super().__init__(message)



#!/usr/bin/env python
# from functools import lru_cache
import json
import logging
import typing as t

import httpx
import xmltodict
from easydict import EasyDict  # type: ignore[import-untyped]

from hut_services.core.cache import file_cache
from hut_services.core.schema import HutSchema
from hut_services.core.schema.geo import BBox
from hut_services.core.service import BaseService
from hut_services.refuges_info.massif import MASSIF_ALPES
from hut_services.refuges_info.schema import (
    RefugesInfoFeatureCollection,
    RefugesInfoHut0Convert,
    RefugesInfoHutSource,
)

if __name__ == "__main__":  # only for testing
    from rich import print as rprint  # noqa: F401, RUF100
    from rich.traceback import install

    install(show_locals=False)


logger = logging.getLogger(__name__)


@file_cache()
def refuges_info_request(
    url: str,
    limit: str | int = "all",
    type_points: t.Sequence[int] = [7, 10, 9, 28],
    massif: t.Sequence[int] | None = MASSIF_ALPES,
    bbox: BBox | None = None,
    text_format: t.Literal["texte", "markdown", "html"] = "markdown",
    output_format: t.Literal["geojson", "xml", "csv"] = "geojson",
    detail: bool = True,
    **params: t.Any,
) -> RefugesInfoFeatureCollection | t.Any | bytes:
    # https://www.refuges.info/api/massif?nb_points=all&format=xml&type_points=7,10,9,28&massif=12,339,407,45,342,20,29,343,412,8,344,408,432,406,52,9
    params["nb_points"] = limit
    params["type_points"] = ",".join([str(t) for t in type_points])
    params["format"] = output_format
    params["format_texte"] = text_format
    params["detail"] = "complet" if detail else "simple"
    if massif and bbox:
        raise NotImplementedError("Either 'massif', or 'bbox' required, both together are not supported.")
    if massif:
        params["massif"] = ",".join([str(m) for m in massif])
        url = url + "/massif"
    elif bbox:
        params["bbox"] = ",".join([str(m) for m in bbox])
        url = url + "/bbox"
    else:
        raise NotImplementedError("Either 'massif', or 'bbox' required.")
    r = httpx.get(url, params=params, timeout=10)
    logger.debug(f"request url: {r.url}")
    if output_format == "geojson":
        data = json.loads(r.content)
        return RefugesInfoFeatureCollection(**data)
    if output_format == "xml":
        return EasyDict(xmltodict.parse(r.content))
    return r.content


class RefugesInfoService(BaseService[RefugesInfoHutSource]):
    """Service to get huts from
    [refuges.info](https://www.refuges.info)
    with [its api](https://www.refuges.info/api).

    Note:
        The methods are descriebed in [`BaseService`][hut_services.BaseService].

    """

    def __init__(self, request_url: str = "https://www.refuges.info/api"):
        super().__init__(support_bbox=True, support_limit=True, support_offset=False, support_convert=True)
        self.request_url = request_url

    def get_huts_from_source(
        self,
        bbox: BBox | None = None,
        limit: int = 1,
        offset: int = 0,
        # type_points: Sequence[int] = [7, 10, 9, 28],
        # massif: Sequence[int] = [12, 339, 407, 45, 342, 20, 29, 343, 412, 8, 344, 408, 432, 406, 52, 9],
        **kwargs: t.Any,
    ) -> list[RefugesInfoHutSource]:
        type_points: t.Sequence[int] = kwargs.get("type_points", [7, 10, 9, 28])
        massif: t.Sequence[int] | None = kwargs.get("massif", MASSIF_ALPES if not bbox else None)
        # "massif", [12, 339, 407, 45, 342, 20, 29, 343, 412, 8, 344, 408, 432, 406, 52, 9] if not bbox else None

        logger.info(f"get refuges.info data from {self.request_url}")
        fc: RefugesInfoFeatureCollection = refuges_info_request(
            url=self.request_url, bbox=bbox, limit=limit, type_points=type_points, massif=massif, detail=True, **kwargs
        )
        huts = []
        for feature in fc.features:
            # rprint(fc)
            refuges_hut = RefugesInfoHutSource(
                name=feature.get_name(),
                source_data=feature,
                source_id=feature.get_id(),
                location=feature.get_location(),
                source_properties=feature.get_properties(),
            )
            huts.append(refuges_hut)
        logger.info(f"succesfully got {len(huts)} huts")
        return huts

    def convert(self, src: t.Mapping | t.Any, include_photos: bool = True) -> HutSchema:
        hut_src = (
            RefugesInfoHutSource(**src)
            if isinstance(src, t.Mapping)
            else RefugesInfoHutSource.model_validate(src, from_attributes=True)
        )
        if hut_src.version >= 0:
            if hut_src.source_data is None:
                err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} without 'source_data' not allowed."
                raise AttributeError(err_msg)
            return RefugesInfoHut0Convert(source_data=hut_src.source_data, include_photos=include_photos).get_hut()
        else:
            err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} not implemented."
            raise NotImplementedError(err_msg)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
    logging.getLogger("chardet").setLevel(logging.WARNING)
    logging.getLogger("tzlocal").setLevel(logging.WARNING)
    limit = 5000
    include_photos = True
    service = RefugesInfoService()
    bbox_ch = None  # (6.02260949059, 45.7769477403, 10.4427014502, 47.8308275417)
    huts = service.get_huts_from_source(limit=limit, bbox=bbox_ch)
    for h in huts:
        # rprint(h)
        hut = service.convert(h, include_photos=include_photos)
        # rprint(h.get_name())
        rprint(hut.name.i18n)
        # rprint(hut.url)
        rprint(hut.photos)
        print("========================")
        # print(hut.description.fr[:250])
        # rprint(h.source_properties_schema)
        # rprint(h)
        # print(h.show(source_name=False))



#!/usr/bin/env python
import datetime
from enum import Enum

from pydantic import Field, computed_field

from ._base import BaseSchema


class ReservationStatusEnum(str, Enum):
    """Enum with reservation status."""

    unknown = "unknown"
    possible = "possible"
    not_possible = "not_possible"
    not_online = "not_online"


class OccupancyStatusEnum(str, Enum):
    """Enum with with occuptation status."""

    unknown = "unknown"
    empty = "empty"
    low = "low"
    medium = "medium"
    high = "high"
    full = "full"

    # unknown = -1
    # empty = 0
    # low = 25
    # medium = 50
    # high = 75
    # full = 100


class PlacesSchema(BaseSchema):
    free: int
    total: int

    @computed_field  # type: ignore[prop-decorator]
    @property
    def occupancy_percent(self) -> float:
        return (self.total - self.free) / self.total * 100 if self.total > 0 else 100

    @computed_field  # type: ignore[prop-decorator]
    @property
    def occupancy_steps(self) -> float:
        steps_fine = round(self.occupancy_percent / 10) * 10
        if self.occupancy_percent > 0 and self.occupancy_percent < 5:
            steps_fine = 10
        elif self.occupancy_percent < 100 and self.occupancy_percent > 95:
            steps_fine = 90
        return steps_fine

    @computed_field  # type: ignore[prop-decorator]
    @property
    def occupancy_status(self) -> OccupancyStatusEnum:
        if self.total == 0:
            return OccupancyStatusEnum.unknown
        if self.occupancy_percent == 100:
            return OccupancyStatusEnum.full
        elif self.occupancy_percent > 62:
            return OccupancyStatusEnum.high
        elif self.occupancy_percent > 37:
            return OccupancyStatusEnum.medium
        elif self.occupancy_percent > 0:
            return OccupancyStatusEnum.low
        return OccupancyStatusEnum.empty


class BookingSchema(BaseSchema):
    date: datetime.date
    reservation_status: ReservationStatusEnum = ReservationStatusEnum.unknown
    unattended: bool
    places: PlacesSchema
    link: str | None = None


class HutBookingsSchema(BaseSchema):
    source_id: int = Field(..., description="ID of the booking provider, e.g. alpsonline.org.")
    start_date: datetime.date
    days: int
    link: str | None = None
    bookings: list[BookingSchema]



# from geojson_pydantic.types import BBox
#
# _ch_lon = [45.7553, 47.6203]
# _ch_lat = [5.7127, 10.5796]
#
# CH: BBox = (_ch_lon[0], _ch_lat[0], _ch_lon[1], _ch_lat[1])

# https://gist.github.com/graydon/11198540



# type: ignore  # noqa: PGH003
import urllib.parse

from icecream import ic
from wikidata.client import Client

## play with wikidata

# qid = "Q42157530"
qid = "Q887175"  # Bluemlisalp
width = 1000

client = Client()  # doctest: +SKIP
entity = client.get(qid, load=True)
ic(entity.description)
image_prop = client.get("P18")
image = entity[image_prop]
ic(entity.attributes)

ic(image.title)
title = urllib.parse.quote(image.title)
link = f"https://commons.wikimedia.org/w/index.php?title=Special:Redirect/file/{title}&width={width}"
ic(link)
ic(image.attributes)
# ic(image.image_resolution)
ic(image.attributes.get("canonicalurl"))



import logging
from enum import Enum
from typing import Annotated

from pydantic import Field, model_validator
from slugify import slugify

from ._base import BaseSchema
from ._contact import ContactSchema
from .locale import TranslationSchema

logger = logging.getLogger(__name__)

NaturalInt = Annotated[int, Field(strict=True, ge=0)]


class OwnerSchema(BaseSchema):
    """Schema for the owner.

    Attributes:
        slug: Owner slug, if empty it is replaced by slugified `name`.
        name: Owner name, required.
        url: Owners URL (not the hut website).
        note: Additonal (public) note to the owner.
        comment: Private comment to the owner.
        contacts: Contacts used for the owner.
    """

    slug: str = Field("", max_length=50)
    name: str = Field(..., max_length=100)
    url: str = Field("", max_length=200)
    note: TranslationSchema = Field(default_factory=TranslationSchema)
    comment: str = ""
    contacts: ContactSchema | None = None

    @model_validator(mode="after")
    def add_slug(self) -> "OwnerSchema":
        if not self.slug:
            self.slug = slugify(self.name, max_length=50, word_boundary=True)
        return self


# T = TypeVar("T")
class HutTypeEnum(str, Enum):
    """Enum with hut types."""

    unknown = "unknown"
    closed = "closed"
    campgr = "campgr"  # campground: possible to camp
    shelter = "shelter"  # basic-shelter: only roof, nothing inside
    camping = "camping"  # attended
    bivouac = "bivouac"  # simple bivouac, not much, high up ...
    selfhut = "selfhut"  # unattended hut
    hut = "hut"
    alp = "alp"
    bhotel = "bhotel"  # simple, not luxiouris hotel, usally with tourist camp
    hostel = "hostel"
    hotel = "hotel"
    special = "special"  # something special, like in a plane or so ...
    resta = "resta"


class CapacitySchema(BaseSchema):
    """Hut capacities.

    Hint:
        For unattended accomodations the `opened` attribute should be used.

    Attributes:
        open:   Capacity when the hut is open
        closed: Capacity when the hut is closed (shelter, winterroom, ...)
    """

    if_open: NaturalInt | None = Field(None, alias="open", description="Capacity when the hut is open")
    if_closed: NaturalInt | None = Field(
        None, alias="closed", description="Capacity when the hut is closed (shelter, winterroom, ...)"
    )


class HutTypeSchema(BaseSchema):
    """Hut type schema.

    Defines the type of the hut if it is open and closed. E.g. a hut can be a 'selfhut' if it is closed.
    If a hut is always open do not add anything to `closed`. If it is specifcally closed, e.g. during winter ass `'closed'`.

    Attributes:
        open:   Type when the hut is open
        closed: Type when the hut is closed (bivouac, selfhut, closed, ...)
    """

    if_open: HutTypeEnum = Field(HutTypeEnum.unknown, alias="open", description="Type when the hut is open")
    if_closed: HutTypeEnum | None = Field(
        None, alias="closed", description="Type when the hut is closed (bivouac, selfhut, closed, ...)"
    )


class PhotoSchemaOld(BaseSchema):
    """Photo schema."""

    attribution: str = Field("", description="Attribution, resp. copyright information as markdown text")
    url: str
    thumb: str = Field("", description="Url to thumbnail (around 400px).")
    caption: TranslationSchema | None = Field(None, description="Original hut name.")
    comment: str = Field("", max_length=20000, description="Additional private comment, e.g for review.")
    is_public: bool = Field(default=True)


class AnswerEnum(str, Enum):
    """Anser enum.

    'yesish' and 'noish' means it is likely to be 'yes' or 'no'.
    'maybe' means is is either 'yes' or 'no'."""

    yes = "yes"
    yesish = "yesish"
    maybe = "maybe"
    noish = "noish"
    no = "no"
    unknown = "unknown"

    # likely_yes = "likely_yes"
    # likely_no = "likely_no"
    # maybe_yes = "maybe_yes"
    # maybe_no = "maybe_no"


class OpenMonthlySchema(BaseSchema):
    """Shows for every month if it is usally, open, partially open or closed.
    Can be accessed as index, but it starts with 1 (month_01)!.

    Attributes:
        url: URL which shows if it is open or not.
        month_mm (OpenMonthly): Month (starting with 01)."""

    url: str = Field("", description="URL which shows if it is open or not.")
    month_01: AnswerEnum = AnswerEnum.unknown
    month_02: AnswerEnum = AnswerEnum.unknown
    month_03: AnswerEnum = AnswerEnum.unknown
    month_04: AnswerEnum = AnswerEnum.unknown
    month_05: AnswerEnum = AnswerEnum.unknown
    month_06: AnswerEnum = AnswerEnum.unknown
    month_07: AnswerEnum = AnswerEnum.unknown
    month_08: AnswerEnum = AnswerEnum.unknown
    month_09: AnswerEnum = AnswerEnum.unknown
    month_10: AnswerEnum = AnswerEnum.unknown
    month_11: AnswerEnum = AnswerEnum.unknown
    month_12: AnswerEnum = AnswerEnum.unknown

    def __getitem__(self, month: int) -> AnswerEnum:
        if month > 0 and month <= 12:
            return getattr(self, f"month_{month:02d}")  # type: ignore[no-any-return]
        else:
            raise IndexError

    def __setitem__(self, month: int, value: AnswerEnum) -> None:
        if month > 0 and month <= 12:
            setattr(self, f"month_{month:02d}", value)
        else:
            raise IndexError

    def __iter__(self):  # type: ignore[no-untyped-def]
        self._current: int = 0
        return self

    def __next__(self) -> AnswerEnum:
        self._current += 1
        if self._current <= 12:
            return self[self._current]
        else:
            raise StopIteration

    def set_month(self, month: int, value: AnswerEnum) -> None:
        if month > 0 and month <= 12:
            setattr(self, f"month_{month:02d}", value)
        else:
            raise IndexError



from typing import Any

import pytest

from hut_services import CapacitySchema, HutTypeEnum, HutTypeSchema
from hut_services.core.guess import guess_hut_type, guess_slug_name


@pytest.mark.parametrize(
    "params,expected",
    [
        (
            {"name": "Waldhuette", "capacity": CapacitySchema(open=40, closed=12), "elevation": 2400.0},
            HutTypeSchema(open=HutTypeEnum.hut, closed=HutTypeEnum.selfhut),
        ),
        (
            {"name": "Berghuette", "capacity": CapacitySchema(open=40, closed=12), "elevation": 3400.0},
            HutTypeSchema(open=HutTypeEnum.hut, closed=HutTypeEnum.bivouac),
        ),
        (
            {
                "name": "Irgend ein Name",
                "capacity": CapacitySchema(open=40, closed=None),
                "elevation": 2000.0,
                "operator": "sac",
            },
            HutTypeSchema(open=HutTypeEnum.hut, closed=None),
        ),
        (
            {"name": "Ohne Beartung", "capacity": CapacitySchema(open=15, closed=None), "operator": "sac"},
            HutTypeSchema(open=HutTypeEnum.selfhut, closed=None),
        ),
        (
            {"name": "Musterbiwak", "capacity": CapacitySchema(open=20, closed=None), "elevation": 2800.0},
            HutTypeSchema(open=HutTypeEnum.bivouac, closed=None),
        ),
        (
            {"name": "Tiefbiwak", "capacity": CapacitySchema(open=20, closed=None), "elevation": 2000.0},
            HutTypeSchema(open=HutTypeEnum.selfhut, closed=None),
        ),
    ],
)
def test_guess_hut_type(params: dict[str, Any], expected: HutTypeSchema) -> None:
    ht = guess_hut_type(**params)
    assert (
        ht.if_open == expected.if_open
    ), f"[{params.get('name')}] open hut type '{ht.if_open}' not as expected '{expected.if_open}'"
    assert (
        ht.if_closed == expected.if_closed
    ), f"[{params.get('name')}] closed hut type '{ht.if_closed}' not as expected '{expected.if_closed}'"


# params: max_length, min_length
@pytest.mark.parametrize(
    "name, params,exp_slug",
    [
        ("Neuste Hütte", {}, "neuste"),
        ("Waldhaus-", {}, "waldhaus"),
        ("I`am a very Long hut name", {"max_length": 15}, "very-long-hut"),
        ("i go to sho t be a wo rd", {}, "i-go-to-sho-t-be-a-wo-rd"),
        ("Name2 with 3numbers5", {}, "name-with-numbers"),
        ("Aarbiwak SAC", {}, "aarbiwak"),
    ],
)
def test_guess_slug(name: str, params: dict[str, Any], exp_slug: str) -> None:
    slug = guess_slug_name(name, **params)
    assert slug == exp_slug, f"Slug not as expcted (name: '{name}')."



import logging
from collections import namedtuple
from collections.abc import Sequence

import phonenumbers
from pydantic import Field

from ._base import BaseSchema
from .locale import TranslationSchema

logger = logging.getLogger(__name__)


PhoneMobile = namedtuple("PhoneMobile", ["phone", "mobile"])


class ContactSchema(BaseSchema):
    """Schema for a contact.

    Attributes:
        name: Contact name (persion or organization), can also be empty.
        email: E-mail address.
        phone: Phone number.
        mobile: Mobule phone number.
        function: Function, e.g. hut warden.
        url: Additional url for this contact (not the hut website).
        address: Address (street, city).
        note: Additional note/information.
        is_active: Contact is active.
        is_public: Show contact public.
    """

    name: str = Field("", max_length=70)
    email: str = Field("", max_length=70)
    phone: str = Field("", max_length=30)
    mobile: str = Field("", max_length=30)
    function: str = Field("", max_length=20)
    url: str = Field("", max_length=200)
    address: str = Field("", max_length=200)
    note: TranslationSchema = Field(default_factory=TranslationSchema)
    is_active: bool = True
    is_public: bool = False

    @classmethod
    def extract_phone_numbers(cls, numbers_string: str, region: str | None) -> list[str]:
        """Extracts phone numbers from a string and returns them formatted
        with international code.
        Uses the [`phonenumbers`](https://github.com/daviddrysdale/python-phonenumbers) package.

        Args:
            numbers_string: A string with phone numbers in it.
            region: Country code.

        Returns:
            A list with formatted phone numbers.
        """
        phones = []
        _matches = phonenumbers.PhoneNumberMatcher(numbers_string, region=region)
        if not _matches:
            logger.warning(f"Could not match phone CH number: '{numbers_string}'")
        phone_match: phonenumbers.PhoneNumberMatch
        for phone_match in _matches:
            phone_fmt = phonenumbers.format_number(phone_match.number, phonenumbers.PhoneNumberFormat.INTERNATIONAL)
            phones.append(phone_fmt)
        return phones

    @classmethod
    def number_to_phone_or_mobile(
        cls, numbers: str | Sequence[str], region: str | None, formatted: bool = False
    ) -> PhoneMobile:
        """Given a phone number it returns it eihter as `phone` or `mobile` number.
        Uses the [`phonenumbers`](https://github.com/daviddrysdale/python-phonenumbers) package.

        Args:
            numbers: List or string of phone numbers, they are extraced and the two assigned to phone and mobile.
            region: Country code.
            formatted: Phone list is alreaad formatted

        Returns:
            Tuple with `phone` and `mobile` number (`(phone, mobile)`).
        """
        numbers_fmt: Sequence[str] = []
        if formatted:
            numbers_fmt = [n.strip() for n in numbers.split(",")] if isinstance(numbers, str) else numbers
        else:
            if not isinstance(numbers, str):
                numbers = " and ".join(numbers)
            numbers_fmt = cls.extract_phone_numbers(numbers, region=region)

        mobiles: list[str] = []
        phones: list[str] = []
        for num in numbers_fmt:
            is_mobile = (
                phonenumbers.number_type(phonenumbers.parse(num, region=region)) == phonenumbers.PhoneNumberType.MOBILE
            )
            if is_mobile:
                mobiles.append(num)
            else:
                phones.append(num)
        mobile = mobiles[0] if mobiles else ""
        phone = phones[0] if phones else ""
        return PhoneMobile(phone=phone, mobile=mobile)



from hut_services import OwnerSchema


def test_set_slug() -> None:
    o = OwnerSchema(name="Test Name")
    assert o.slug == "test-name"


def test_set_slug_special_name() -> None:
    o = OwnerSchema(name="Test Name ... #@")
    assert o.slug == "test-name"


def test_set_slug_with_long_name() -> None:
    o = OwnerSchema(name="Test Name so long that it cannot stop to give some shorter Slug.")
    assert o.slug == "test-name-so-long-that-it-cannot-stop-to-give-some"


def test_with_slug() -> None:
    o = OwnerSchema(slug="my-slug", name="Test Name")
    assert o.slug == "my-slug"



import logging
import time
from io import BytesIO

import dateparser
import requests
from bs4 import BeautifulSoup
from PIL import ImageFile
from pydantic_string_url import HttpUrl

# if __name__ == "__main__":
from rich import print as rprint

# from numpy import imag
from hut_services.core.cache import file_cache
from hut_services.core.schema._license import AuthorSchema, LicenseSchema, SourceSchema
from hut_services.core.schema._photo import PhotoSchema
from hut_services.core.schema.locale import TranslationSchema

logging.getLogger("PIL").setLevel(logging.WARNING)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("chardet").setLevel(logging.WARNING)
logging.getLogger("tzlocal").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

refuges_lic = LicenseSchema(
    slug="cc-by-sa-2.0", name="CC-BY-SA 2.0", url="https://creativecommons.org/licenses/by-sa/2.0/"
)


@file_cache(forever=True)
def _get_image_size(url: HttpUrl, _delay: float = 0.2) -> tuple[int, int]:
    response = requests.get(url, stream=True, timeout=15)
    image_data = BytesIO()
    image_file = ImageFile.Parser()
    time.sleep(_delay)
    for counter, chunk in enumerate(response.iter_content(4096)):
        image_data.write(chunk)
        image_file.feed(chunk)
        if image_file.image:
            logger.debug(f"Image size determined after {counter+1} iterations ({url})")
            return image_file.image.size
    return 0, 0


@file_cache(forever=True)
def _get_original_images_request(hut_id: str, _delay: float = 1.5) -> bytes:
    url = f"https://www.refuges.info/point/{hut_id}"
    response = requests.get(url, timeout=15)
    time.sleep(_delay)
    return response.content


@file_cache()
def get_original_images(hut_id: str) -> list[PhotoSchema]:
    soup = BeautifulSoup(_get_original_images_request(hut_id), "html.parser")
    comments = soup.find_all("li")
    original_images = []
    for comment in comments:
        image_tag = comment.find("img")
        if image_tag:
            image_url = f"https://www.refuges.info{image_tag.parent['href']}".split("?")[0]
            image_url = image_url.replace("-reduite", "-originale")
            fauxfieldset_legend_split = comment.find("p", class_="fauxfieldset-legend").text.split("par")
            capture_date_str_fr = fauxfieldset_legend_split[0].strip()
            _capture_date_str_image = comment.find("div", class_="texte_sur_image").text  # date in the image - not used

            try:
                capture_date = dateparser.parse(capture_date_str_fr, languages=["fr"])
            except ValueError:
                rprint(comment)
                rprint(capture_date_str_fr)
                rprint(_capture_date_str_image)
                rprint(f"https://www.refuges.info/point/{hut_id}")
                raise
            try:
                caption = TranslationSchema(
                    fr=comment.find("blockquote").text.strip(),
                )
            except AttributeError:
                # no comment to the image
                caption = TranslationSchema()
            author_name = fauxfieldset_legend_split[1].strip() if len(fauxfieldset_legend_split) > 1 else ""
            author = AuthorSchema(name=author_name) if author_name else None
            src_ident = comment.find("a")["id"]
            src_url = f"https://www.refuges.info/point/{hut_id}#{src_ident}"
            source = SourceSchema(name="refuges.info", url=src_url, ident=src_ident)
            width, height = 0, 0  # _get_image_size(image_url)
            photo_schema = PhotoSchema(
                raw_url=image_url,
                url=src_url,
                capture_date=capture_date,
                caption=caption,
                comment="",
                author=author,
                source=source,
                width=width,
                height=height,
                licenses=[refuges_lic],
            )
            original_images.append(photo_schema)
    return original_images


if __name__ == "__main__":
    photos = get_original_images("9819")
    rprint(photos)



from typing import Annotated, TypeAlias

from geojson_pydantic.types import BBox, Position  # noqa: F401
from pydantic import Field

# WSG84
Elevation = Annotated[float, Field(ge=0, lt=9000, description="Elevation in meter")]
Longitude = Annotated[float, Field(ge=-180, le=180, examples=["7.6496971"], title="Longitude (x) in WGS84")]
Latitude = Annotated[float, Field(ge=-90, le=90, examples=["45.9765729"], title="Latitude (y) in WGS84")]

# LV03 (swiss)
LV03X = Annotated[float, Field()]  # ge=-90, le=90, examples=["45.9765729"], title="Latitude (y) in WGS84")]
LV03Y = Annotated[float, Field()]  # ge=-90, le=90, examples=["45.9765729"], title="Latitude (y) in WGS84")]
LV03: TypeAlias = tuple[LV03Y, LV03X, Elevation]



import logging

from pydantic import Field, computed_field

from hut_services import (
    BaseHutConverterSchema,
    BaseHutSourceSchema,
    LocationEleSchema,
    SourceDataSchema,
    SourcePropertiesSchema,
    TranslationSchema,
)
from hut_services.core.schema.geo import BBox
from hut_services.core.schema.geo.types import Latitude, Longitude
from hut_services.osm.schema import OSMTagsOptional

# from hut_services.wikidata.service import WikidataEntity  # , wikidata_service


logger = logging.getLogger(__name__)


class GeocodeHutSchema(SourceDataSchema):
    """Open street map schema."""

    place_id: int | None = None
    name: str
    lat: Latitude
    lon: Longitude
    licence: str | None = None
    osm_type: str | None = None
    osm_id: int | None = None
    category: str | None = None
    building_type: str | None = Field(None, alias="type")
    place_rank: int | None = None
    importance: float | None = None
    addresstype: str | None = None
    display_name: str | None = None
    extratags: OSMTagsOptional | None = None
    boundingbox: BBox | None = None  # ['47.0545528', '47.0548247', '11.1981883', '11.1984526']

    def get_id(self) -> str:
        """Get open street map `id`."""
        return str(self.place_id)

    def get_name(self) -> str:
        """Get open street map hut name."""
        return self.name

    def get_location(self) -> LocationEleSchema:
        """Get open street map location."""
        _tags = self.extratags if self.extratags else OSMTagsOptional()  # pyright: ignore  # noqa: PGH003
        try:
            ele = float(_tags.ele)  # type: ignore  # noqa: PGH003
        except (ValueError, TypeError):
            ele = None
        return LocationEleSchema(lat=self.lat, lon=self.lon, ele=ele)


class GeocodeProperties(SourcePropertiesSchema):
    pass


class GeocodeHutSource(BaseHutSourceSchema[GeocodeHutSchema, GeocodeProperties]):
    version: int = 0
    source_name: str = "geocode"


class GeocodeHut0Convert(BaseHutConverterSchema[GeocodeHutSchema]):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def name(self) -> TranslationSchema:
        return TranslationSchema(de=self.source_data.get_name())

    @computed_field  # type: ignore[prop-decorator]
    @property
    def description(self) -> TranslationSchema:
        return TranslationSchema()



from tasks import (
    Ctx,
    doc,
    header,
    success,
    task,
)


@task
def lock(c: Ctx):
    """Checking uv lock file"""
    header(doc())
    c.run("uv lock --locked", echo=True, pty=True)


@task
def lint(c: Ctx):
    """Linting code"""
    header(doc())
    c.run("uv run pre-commit run -a", echo=True, pty=True)


@task
def fix(c: Ctx):
    """Fix code with ruff"""
    header(doc())
    c.run("uv run ruff check --fix .", echo=True, pty=True)


@task
def deps(c: Ctx):
    """Checking dependencies"""
    header(doc())
    c.run("uv run deptry .", echo=True, pty=True)


@task
def types(c: Ctx):
    """Static type checking"""
    header(doc())
    c.run("uv run mypy ", echo=True, pty=True)
    # c.run("uv run pyright  src", echo=True, pty=True)


@task
def test(c: Ctx):
    """Static type checking"""
    header(doc())
    c.run("pytest -v", echo=True, pty=True)


@task(pre=[lock, lint, deps, types], default=True)
def check(c: Ctx):
    """Run code quality tools"""
    header("Summary")
    success("Code quality checks passed")



import pytest

from hut_services.core.schema import HutSchema
from hut_services.osm import OsmService
from hut_services.osm.schema import OsmHutSource

HUT_LIMIT: int = 2


@pytest.fixture(scope="module")
def service() -> OsmService:
    return OsmService()


@pytest.fixture(scope="module")
def hut_sources(service) -> list[OsmHutSource]:
    return service.get_huts_from_source(limit=HUT_LIMIT)


@pytest.fixture(scope="module")
def huts(service) -> list[HutSchema]:
    return service.get_huts(limit=HUT_LIMIT)


def test_osm_service_source_online(hut_sources: list[OsmHutSource]) -> None:
    """Get huts from opemstreemap.org"""
    assert len(hut_sources) == HUT_LIMIT
    for h in hut_sources:
        # rprint(h)
        print(h.show(source_name=False))
        assert type(h) is OsmHutSource


def test_osm_service_hut_online(huts: list[HutSchema]) -> None:
    """Tests conversion to HutSchema as well."""
    assert len(huts) == 2
    assert type(huts[0]) is HutSchema


def test_osm_service_convert_dict_online(hut_sources: list[OsmHutSource], service: OsmService) -> None:
    """Convert by a dict"""
    for h in hut_sources:
        h_dict = h.model_dump(by_alias=True)
        # add a field
        h_dict["psuedo"] = "No needed"
        service.convert(h_dict)


def test_osm_service_convert_obj_online(hut_sources: list[OsmHutSource], service: OsmService) -> None:
    """Convert by a random class with missing location"""

    class MySource:
        def __init__(self, data: dict, name: str):
            self.source_data = data
            self.source_id = 10
            self.name = name
            self.random = 200

    for h in hut_sources:
        h_obj = MySource(data=h.source_data.model_dump(by_alias=True) if h.source_data else {}, name="MyName")
        service.convert(h_obj)


def test_osm_service_convert_obj2_online(hut_sources: list[OsmHutSource], service: OsmService) -> None:
    """Convert by a random class with location (x,y,z)"""

    class MySource:
        def __init__(self, data: dict, name: str):
            self.source_data = data
            self.source_id = 10
            self.location: dict = {"x": 45, "y": 10, "z": 2000}
            self.name = name
            self.random = 200

    for h in hut_sources:
        h_obj = MySource(data=h.source_data.model_dump(by_alias=True) if h.source_data else {}, name="MyName")
        h_c = service.convert(h_obj)
        assert h_c.name.i18n == h.name



import re
from typing import Literal

from slugify import slugify

__all__ = ["guess_hut_type", "guess_slug_name"]

from hut_services.core.schema._hut_fields import (
    AnswerEnum,
    CapacitySchema,
    HutTypeEnum,
    HutTypeSchema,
    OpenMonthlySchema,
)

HUT_NAMES = [r"huette", r"r[ie]fug[ei]", r"h[iü]tt[ae]", r"camona", r"capanna", r"cabane", r"huisli"]
BIVI_NAMES = [r"biwak", r"bivouac", r"bivacco"]
BASIC_HOTEL_NAMES = [r"berghotel", r"berggasthaus", r"auberge", r"gasthaus", r"berghaus"]
CAMPING_NAMES = [r"camping", r"zelt"]
HOTEL_NAMES = [r"h[oô]tel"]
HOSTEL_NAMES = [r"hostel", r"jugendherberg"]
RESTAURANT_NAMES = [r"restaurant", r"ristorante", r"beizli"]
_ALP_NAMES = ["alp", "alm", "hof"]


def _in(patterns: list, target: str) -> bool:
    return any(re.search(pat.lower(), target.lower()) for pat in patterns)


def guess_hut_type(
    name: str = "",
    default: HutTypeEnum = HutTypeEnum.unknown,
    capacity: CapacitySchema | None = None,
    elevation: float | None = 1500,
    operator: Literal["sac", "dav"] | None = None,
    osm_tag: str | None = "",
    missing_walls: int | None | str = 0,
    open_monthly: OpenMonthlySchema | None = None,
    # ) -> HutType:
) -> HutTypeSchema:
    """Guess hut type based on some input parameters.

    Args:
        name: hut name
        default: default type if nothing else fits
        capacity: capacity for a open and closed hut
        operator: who is operating the hut
        osm_tag: osm toursm tag
        missing_walls: missing_walls value from refuges.info
        open_monthly: list which month it is open"""
    # check if every month is closed
    is_closed = False if open_monthly is None else all(o == AnswerEnum.no for o in open_monthly)
    # if capacity is not None and capacity.if_open == 0 and capacity.if_closed in (0, None):
    #    is_closed = True

    name = name or ""
    capacity_open = capacity.if_open or 0 if capacity is not None else 0
    capacity_closed = capacity.if_closed or 0 if capacity is not None else 0
    elevation = elevation or 1500
    osm_tag = osm_tag or ""
    missing_walls = missing_walls or 0
    if isinstance(missing_walls, str):
        try:
            missing_walls = int(missing_walls)
        except ValueError:
            missing_walls = 0

    name = name.lower()
    _possible_hut = _in(HUT_NAMES, name)
    slug_open = default
    if is_closed:
        slug_open = HutTypeEnum.closed
    elif _in(BASIC_HOTEL_NAMES, name):
        slug_open = HutTypeEnum.bhotel
    elif _in(HOTEL_NAMES, name):
        slug_open = HutTypeEnum.hotel
    elif _in(HOSTEL_NAMES, name):
        slug_open = HutTypeEnum.hostel
    elif _in(RESTAURANT_NAMES, name):
        slug_open = HutTypeEnum.resta
    elif _in(CAMPING_NAMES, name):
        slug_open = HutTypeEnum.camping
    elif osm_tag == "wilderness_hut" or missing_walls > 0:
        slug_open = HutTypeEnum.bivouac if elevation > 2500 and not _possible_hut else HutTypeEnum.shelter
    elif (capacity_open == capacity_closed or capacity_open < 22) and capacity_open > 0:
        slug_open = HutTypeEnum.bivouac if elevation > 2500 and not _possible_hut else HutTypeEnum.selfhut
    elif _possible_hut:
        slug_open = HutTypeEnum.hut
    elif _in(BIVI_NAMES, name):
        slug_open = HutTypeEnum.selfhut if elevation < 2200 else HutTypeEnum.bivouac
    elif _in(_ALP_NAMES, name) and elevation < 2000:
        slug_open = HutTypeEnum.alp
    elif operator in ["sac", "dav"] or osm_tag == "alpine_hut":
        slug_open = HutTypeEnum.hut
    slug_closed = None
    if slug_open == HutTypeEnum.hut and capacity_open > 0 and capacity_open != capacity_closed:
        if (capacity is not None and capacity.if_closed is not None) and capacity.if_closed == 0:
            slug_closed = HutTypeEnum.closed
        elif (capacity is not None and capacity.if_closed is not None) and capacity.if_closed > 0:
            slug_closed = HutTypeEnum.selfhut if elevation < 3000 else HutTypeEnum.bivouac
    return HutTypeSchema(open=slug_open, closed=slug_closed)


def guess_slug_name(hut_name: str, max_length: int = 25, min_length: int = 4) -> str:
    REPLACE_IN_SLUG = [
        "alpage",
        "alpina",
        "huette",
        "cabanne",
        "cabane",
        "capanna",
        "chamana",
        "chamanna",
        "chamonna",
        "chalet",
        "capanna",
        "biwak",
        "bivouac",
        "bivacco",
        "bergerie",
        "forestiere",
        "pastorale",
        "baraque",
        "berghotel",
        "chalets",
        "camona",
        "hotel",
        "forestier",
        "chapelle",
        "chambres",
        "fort",
        "schloss",
        "grange",
        "habert",
        "hameau",
        "maison",
        "malga",
        "hote",
        "cabanes",
        "camping",
        "casera",
        "casina",
        "centre",
        "commune",
        "communal",
        "casere",
        "cantine",
        "huette",
        "naturfreundehaus",
        "naturfreunde",
        "berghuette",
        "berggasthaus",
        "waldhuette",
        "berghaus",
        "cascina",
        "rifugio",
        "restaurant",
        "ricovero",
        "refuge",
        "citta",
        "guide",
        "zollhuette",
    ]
    NOT_IN_SLUG = [
        *REPLACE_IN_SLUG,
        "alp",
        "alpe",
        "la",
        "le",
        "gite",
        "casa",
        "sac",
        "cas",
        "caf",
        "cai",
        "del",
        "des",
        "les",
        "sous",
        "rif",
        "abri",
        "sur",
        "ski",
        "aacz",
        "aacb",
    ]

    for r in ("ä", "ae"), ("ü", "ue"), ("ö", "oe"):
        hut_name = hut_name.lower().replace(r[0], r[1])
    slug = slugify(hut_name)
    slug = re.sub(r"[0-9]", "", slug)  # remove numbers
    slug = slug.strip(" -")
    slugs = slug.split("-")
    slugl = [s for s in slugs if (s not in NOT_IN_SLUG and len(s) >= 3)]
    for _replace in REPLACE_IN_SLUG:
        slugl = [
            v.replace(_replace, "") if len(v.replace(_replace, "")) > 4 else v for v in slugl if v.replace(_replace, "")
        ]
    if not slugl or len("-".join(slugl)) < min_length:
        slugl = slugify(hut_name).split("-")
    return slugify(" ".join(slugl), max_length=max_length, word_boundary=True)



import logging
import textwrap
from typing import Any, Literal

from pydantic import BaseModel, Field, computed_field

from hut_services import (
    BaseHutConverterSchema,
    BaseHutSourceSchema,
    CapacitySchema,
    ContactSchema,
    HutTypeEnum,
    HutTypeSchema,
    LocationEleSchema,
    OwnerSchema,
    PhotoSchema,
    SourceDataSchema,
    SourcePropertiesSchema,
    TranslationSchema,
)
from hut_services.core.schema.geo.types import Elevation, Latitude, Longitude

from ..core.guess import guess_hut_type

# from hut_services.wikidata import WikidataService
# from hut_services.wikidata.service import WikidataEntity  # , wikidata_service
from .exceptions import OSMCoordinatesError

logger = logging.getLogger(__name__)


class OSMTags(BaseModel):
    """Open street map tags."""

    tourism: Literal["alpine_hut", "wilderness_hut"]
    wikidata: str | None = None

    name: str
    operator: str | None = None
    email: str | None = None
    contact_email: str | None = Field(None, alias="contact:field")
    phone: str | None = None
    contact_phone: str | None = Field(None, alias="contact:phone")
    website: str | None = None
    contact_website: str | None = Field(None, alias="contact:website")
    note: str | None = None

    bed: str | None = None
    beds: str | None = None
    capacity: str | None = None
    access: str | None = None
    fireplace: str | None = None
    wall: str | None = None
    amenity: str | None = None
    shelter_type: str | None = None
    winter_room: str | None = None
    reservation: str | None = None

    # ele: Optional[Elevation]
    ele: Elevation | None = None


class OSMTagsOptional(OSMTags):
    """Open street map tags, all optional."""

    tourism: str | None = None  # type: ignore[assignment]
    name: str | None = None  # type: ignore[assignment]
    ele: float | str | None = None  # type: ignore[assignment]


class OsmHutSchema(SourceDataSchema):
    """Open street map schema."""

    osm_type: Literal["node", "way", "area"] | None = None
    osm_id: int = Field(..., alias="id")
    lat: Latitude | None = None
    lon: Longitude | None = None
    center_lat: Latitude | None = None
    center_lon: Longitude | None = None
    tags: OSMTags

    def get_id(self) -> str:
        """Get open street map `id`."""
        return str(self.osm_id)

    def get_name(self) -> str:
        """Get open street map hut name."""
        return self.tags.name

    def get_location(self) -> LocationEleSchema:
        """Get open street map location."""
        lat: float | None
        lon: float | None
        # coordinate fixes
        if self.osm_id == 1386596359:  # Winteregghütte
            lat, lon = 46.46039, 7.64995
        elif self.osm_id == 505804029:  # Ostegg
            lat, lon = 46.60037, 8.04120
        else:
            lat, lon = self.lat, self.lon
        if lat and lon:
            return LocationEleSchema(lat=lat, lon=lon, ele=self.tags.ele)
        elif self.center_lat and self.center_lon:
            return LocationEleSchema(lat=self.center_lat, lon=self.center_lon, ele=self.tags.ele)
        else:
            raise OSMCoordinatesError(self.osm_id, self.get_name())


class OsmProperties(SourcePropertiesSchema):
    osm_type: Literal["node", "way", "area"] = Field(..., description="osm object type: node, way, or area")


class OsmHutSource(BaseHutSourceSchema[OsmHutSchema, OsmProperties]):
    version: int = 0
    source_name: str = "osm"


class OsmHut0Convert(BaseHutConverterSchema[OsmHutSchema]):
    include_photos: bool = False

    @property
    def _tags(self) -> OSMTags:
        return self.source_data.tags

    # implemented in base
    # @computed_field  # type: ignore[prop-decorator]
    # @property
    # def slug(self) -> str:
    #    return f"osm-{self.source.get_id()}"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def name(self) -> TranslationSchema:
        return TranslationSchema(de=self._tags.name[:69])

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source_name(self) -> str:
        return "osm"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def description(self) -> TranslationSchema:
        return TranslationSchema()

    @computed_field  # type: ignore[prop-decorator]
    @property
    def owner(self) -> OwnerSchema | None:
        name = self._tags.operator or ""
        comment = ""
        if len(name) > 60:
            comment = f"Full name: {name}"
            name = textwrap.shorten(name, width=57, placeholder="...")
        if name:
            return OwnerSchema(name=name, comment=comment)
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def url(self) -> str:
        url = ""
        if self._tags.website:
            url = self._tags.website
        elif self._tags.contact_website:
            url = self._tags.contact_website
        if len(url) > 200:
            url = ""
        return url

    ## Contact
    @property
    def _email(self) -> str:
        if self._tags.email:
            return self._tags.email.strip()
        elif self._tags.contact_email:
            return self._tags.contact_email.strip()
        return ""

    @property
    def _phones(self) -> list[str]:
        phone = None
        if self._tags.phone:
            phone = self._tags.phone
        elif self._tags.contact_phone:
            phone = self._tags.contact_phone
        phones = []
        if phone:
            phones += ContactSchema.extract_phone_numbers(phone, region="CH")
        return phones

    @computed_field  # type: ignore[prop-decorator]
    @property
    def contacts(self) -> list[ContactSchema]:
        contacts = []
        emails = self._email
        for phone in self._phones:
            phone, mobile = ContactSchema.number_to_phone_or_mobile(phone, region="CH", formatted=True)
            contacts.append(
                ContactSchema(
                    phone=phone, email=emails.strip(), mobile=mobile, name="", function="contact", is_public=True
                )
            )
            if emails:
                emails = ""
        if emails:
            for email in emails.split(";"):
                contacts.append(ContactSchema(email=email.strip(), function="contact", is_public=True))
        return contacts

    @computed_field  # type: ignore[prop-decorator]
    @property
    def comment(self) -> str:
        note = ""
        if self._tags.note:
            note = f"OSM note: {self._tags.note}\n"
        return note

    @property
    def _capacity_opened(self) -> int | None:
        tags = self._tags
        cap: int | None = None
        cap_str: str | None = None
        if tags.capacity:
            cap_str = tags.capacity
        elif tags.beds:
            cap_str = tags.beds
        elif tags.bed:
            cap_str = tags.bed
        try:
            if cap_str is not None:
                cap = int(cap_str)
        except ValueError:
            cap = None
        return cap

    @property
    def _capacity_closed(self) -> int | None:
        if self._tags.winter_room:
            number = None
            try:
                number = int(self._tags.winter_room)  # capacity in tag
            except ValueError:
                return None
            return number
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def capacity(self) -> CapacitySchema:
        closed = self._capacity_closed
        if closed == self._capacity_opened:
            closed = None
        return CapacitySchema(open=self._capacity_opened, closed=closed)

    @computed_field(alias="type")  # type: ignore[prop-decorator]
    @property
    def hut_type(self) -> HutTypeSchema:
        capacity = CapacitySchema(open=self._capacity_opened, closed=self._capacity_closed)
        hut_types = guess_hut_type(
            name=self.name.i18n or "",
            capacity=capacity,
            elevation=self.location.ele,
            operator="sac" if self._tags.operator and "sac" in self._tags.operator else None,
            osm_tag=self._tags.tourism,
        )
        if hut_types.if_open == HutTypeEnum.hut and capacity.if_open in [0, None]:
            hut_types.if_open = HutTypeEnum.unknown
            hut_types.if_closed = None
        return hut_types

    @property
    def wikidata_entity(self) -> None:
        # if self._tags.wikidata and self.get_wikidata_photos:
        #    return wikidata_service.get_entity(self._tags.wikidata)
        return None

    @computed_field()  # type: ignore[prop-decorator]
    @property
    def photos(self) -> list[PhotoSchema]:
        """Not supported for osm data, use wikidata isntead which are based on the osm data."""
        if self.wikidata_entity is not None and self.include_photos:
            return self.wikidata_entity.get_photos()
        return []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_active(self) -> bool:
        return True

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_public(self) -> bool:
        if self._tags.access:
            return self._tags.access in ["yes", "public", "customers", "permissive"]
        return True

    @computed_field  # type: ignore[prop-decorator]
    @property
    def extras(self) -> dict[str, Any]:
        extras = {}
        if self._tags.wikidata:
            extras["wikidata"] = self._tags.wikidata
        return extras



from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field, computed_field

from ..guess import guess_slug_name
from ._contact import ContactSchema
from ._hut import HutSchema
from ._hut_base_source import SourceDataSchema
from ._hut_fields import (
    CapacitySchema,
    HutTypeEnum,
    HutTypeSchema,
    OpenMonthlySchema,
    OwnerSchema,
)
from ._license import AuthorSchema, SourceSchema
from ._photo import PhotoSchema
from .geo import LocationEleSchema
from .locale import TranslationSchema

if __name__ == "__main__":  # only for testing
    from icecream import ic  # type: ignore[import-untyped] # noqa: F401, RUF100 , PGH003


TSourceData = TypeVar("TSourceData", bound=SourceDataSchema)


class BaseHutConverterSchema(BaseModel, Generic[TSourceData]):
    """Base class used for a converter schema.

    Attributes:
        source: Source data

    All attributes as in [`HutSchema`][hut_services.core.schema.HutSchema] either as
    attribute or pydantic `computed_field`.

    Examples:
        See `hut_services.osm.schema.OsmHut0Convert`."""

    # See [`OsmHut0Convert`][hut_services.osm.schema.OsmHut0Convert].

    source_data: TSourceData = Field(..., exclude=True)
    include_photos: bool = True

    class FieldNotImplementedError(Exception):
        """Field is not implemented.

        Args:
            obj: Current object.
            field: Field which is not implementd."""

        def __init__(self, obj: "BaseHutConverterSchema", field: str):
            class_name = str(obj.__class__).replace("<class '", "").replace("'>", "")
            message = f"Converter '{class_name}' field '{field}' is not implemented."
            super().__init__(message)

    def get_hut(self) -> HutSchema:
        """Convert to hut.

        Returns:
            Converted hut."""
        hut_dict = self.model_dump(by_alias=True)
        return HutSchema(**hut_dict)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def slug(self) -> str:
        return guess_slug_name(self.name.i18n)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def name(self) -> TranslationSchema:
        if hasattr(self.source_data, "get_name"):
            return TranslationSchema(de=self.source_data.get_name())
        raise self.FieldNotImplementedError(self, "name")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source(self) -> SourceSchema | None:
        if hasattr(self.source_data, "get_name"):
            return SourceSchema(name=self.source_name, ident=self.source_data.get_id())
        raise self.FieldNotImplementedError(self, "origin")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def author(self) -> AuthorSchema | None:
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source_name(self) -> str:
        # TODO: how to get source_name
        # if hasattr(self.source_data, "source_name"):
        #     return self.source_data.source_name
        raise self.FieldNotImplementedError(self, "source_name")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def location(self) -> LocationEleSchema:
        if hasattr(self.source_data, "get_location"):
            return self.source_data.get_location()  # type: ignore  # noqa: PGH003
        raise self.FieldNotImplementedError(self, "location")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def description(self) -> TranslationSchema:
        raise self.FieldNotImplementedError(self, "description")

    # @computed_field  # type: ignore[prop-decorator]
    # @property
    # def description_attribution(self) -> str:
    #    return ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def notes(self) -> list[TranslationSchema]:
        return []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def owner(self) -> OwnerSchema | None:
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def url(self) -> str:
        return ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def contacts(self) -> list[ContactSchema]:
        return []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def country_code(self) -> str | None:
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def comment(self) -> str:
        return ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def capacity(self) -> CapacitySchema:
        return CapacitySchema(open=None, closed=None)

    @computed_field(alias="type")  # type: ignore[prop-decorator]
    @property
    def hut_type(self) -> HutTypeSchema:
        return HutTypeSchema(open=HutTypeEnum.unknown, closed=None)

    @property
    def photos(self) -> list[PhotoSchema]:
        return []

    @computed_field()  # type: ignore[prop-decorator]
    @property
    def open_monthly(self) -> OpenMonthlySchema:
        return OpenMonthlySchema()  # pyright: ignore  # noqa: PGH003

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_active(self) -> bool:
        return True

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_public(self) -> bool:
        return True

    @computed_field  # type: ignore[prop-decorator]
    @property
    def extras(self) -> dict[str, Any]:
        return {}



import os
import tempfile
from collections.abc import Callable, Sequence
from typing import Any, TypeVar

from joblib import Memory, expires_after  # type: ignore[import-untyped]

__all__ = ["clear_file_cache", "file_cache"]


cachedir = os.environ.get("HUT_SERVICE_CACHE_DIR", os.path.join(tempfile.gettempdir(), "py_file_cache"))
default_seconds = int(os.environ.get("HUT_SERVICE_EXPIRE_SECONDS", 3600 * 24 * 2))  # 2 days
forever_seconds = int(3600 * 24 * 365 * 10)  # 10 years
_memory = Memory(cachedir, verbose=0, compress=True)
T = TypeVar("T")


def file_cache(
    func: None | Callable = None,
    ignore: Sequence = [],
    expire_in_seconds: int = default_seconds,
    forever: bool = False,
) -> Any:
    """Chaches function, if 'forever' is True, cache will never expire (and 'expire_in_seconds' will be ignored)."""
    if forever:
        expire_in_seconds = forever_seconds
    return _memory.cache(func=func, ignore=ignore, cache_validation_callback=expires_after(seconds=expire_in_seconds))


clear_file_cache = _memory.clear
"""Cleares cache."""



import inspect
import shutil
import sys

from environs import Env
from rich import print as rprint

env = Env()
env.read_env()  # read .env file, if it exists


def echo(*args, raw=False, **kwargs):
    if raw:
        print(*args, **kwargs)
    else:
        rprint(*args, **kwargs)


def header(msg: str = "", symb="-", style="blue bold", max_length=None, stderr: bool = True):
    terminal_width = shutil.get_terminal_size(fallback=(80, 20)).columns
    max_length = int(terminal_width)
    max_length = 50 if max_length < 50 else 120 if max_length > 120 else max_length
    min_ = 6
    max_ = max_length - min_ - 2
    length = len(msg)
    start = f"[dim]{symb * (max_ - length) if length < max_ else symb * min_}[/]"
    end = f"[dim]{symb * min_}[/]"
    if not msg:
        echo(f"[dim]{symb * max_length}[/]", file=sys.stderr if stderr else None)
    else:
        echo(f"{start} [{style}]{msg} {end}", file=sys.stderr if stderr else None)


def info(msg: str, stderr: bool = True):
    echo("[blue]Info   :[/]", msg, file=sys.stderr if stderr else None)


def success(msg: str, stderr: bool = True):
    echo("[green bold]Success:[/]", msg, file=sys.stderr if stderr else None)


def warning(msg: str, stderr: bool = True):
    echo("[yellow]Warning:[/]", msg, file=sys.stderr if stderr else None)


def error(msg: str, stderr: bool = True):
    echo(f"[red bold]Error:  [/] [red]{msg}[/]", file=sys.stderr if stderr else None)
    sys.exit(1)


def doc():
    """Retrieve and print the docstring of the caller function."""
    caller_frame = inspect.stack()[1]
    caller = caller_frame.frame.f_globals.get(caller_frame.function)
    return inspect.getdoc(caller) or ""



import logging

from pydantic import Field
from pydantic_string_url import HttpUrl

from ._base import BaseSchema

logger = logging.getLogger(__name__)


class LicenseSchema(BaseSchema):
    """
    License information.

    Attributes:
        slug: Identifier or slug of the license (e.g. by-sa, cc-0, ...).
        url: URL of the license.
        name: Fullname of the license.
    """

    slug: str
    url: HttpUrl | None = None
    name: str = Field("", max_length=300)


class AuthorSchema(BaseSchema):
    """
    Author information.

    Attributes:
        name: Name of the author.
        url: URL of the author.
    """

    name: str
    url: None | HttpUrl = None


class SourceSchema(BaseSchema):
    """
    Source information.

    Attributes:
        ident: Identifier or slug of the source.
        name: Name of the source, either person or organization (e.g refuges, wikicommon, sac, ...)
        url: URL of the source.
    """

    ident: str = Field("", max_length=300)
    name: str = Field("", max_length=300)
    url: HttpUrl | None = None



#!/usr/bin/env python
# from functools import lru_cache
import logging
import typing as t

# from typing import Any, Literal, Mapping
from wikidata.client import Client
from wikidata.entity import EntityId

from hut_services import BaseService, file_cache
from hut_services.core.schema import HutSchema
from hut_services.core.schema.geo import BBox
from hut_services.osm.service import OsmService
from hut_services.wikidata.schema import (
    WikidataHut0Convert,
    WikidataHutSchema,
    WikidataHutSource,
    WikidataPhoto,
    WikidataProperties,
)

if __name__ == "__main__":  # only for testing
    from icecream import ic  # type: ignore[import-untyped] # noqa: F401, RUF100 , PGH003
    from rich import print as rprint  # noqa: F401, RUF100
    from rich.traceback import install

    install(show_locals=False)


logger = logging.getLogger(__name__)


@file_cache(ignore=["client"])
def _get_photo(client: Client, qid: EntityId) -> WikidataPhoto | None:
    image_prop = client.get(EntityId("P18"))  # image
    entity = client.get(qid, load=True)
    wikidata_url = f"https://www.wikidata.org/wiki/{qid.upper()}"
    if image_prop not in entity:
        logger.debug(f"No wikidata image for: '{wikidata_url}'")
        return None
    image = entity[image_prop]
    logger.info(f"Got wikidata image entity: '{image.title}'")  # type: ignore[attr-defined]
    return WikidataPhoto.model_validate(image)


@file_cache(ignore=["client"])
def _get_attributes(client: Client, qid: EntityId) -> dict[str, t.Any]:
    entity = client.get(qid, load=True)
    return dict(entity.attributes.items())


class WikidataEntity:
    def __init__(self, qid: EntityId, client: Client | None):
        self.client = Client() if client is None else client
        self.qid = qid

    def get_photo(self) -> WikidataPhoto | None:
        """Get a list of photos from wikidata."""
        photo: WikidataPhoto | None = _get_photo(client=self.client, qid=self.qid)
        return photo

    def get_attributes(
        self,
    ) -> dict[str, t.Any]:
        """Get a list of photos from wikidata."""
        attributes = _get_attributes(client=self.client, qid=self.qid)
        # assert all(
        #    isinstance(p, dict[str, t.Any]) for p in attributes
        # ), "Wrong type, not a list of 'PhotoSchema'"
        return t.cast(dict[str, t.Any], attributes)


class WikidataService(BaseService[WikidataHutSource]):
    """Service to get Information from
    [Wikidata](https://www.wikidata.org/)
    with a [Python Wikidata](https://pypi.org/project/Wikidata/) client.

    Note:
        It is not (yet) possible to get huts and convert them.
    """

    def __init__(self, request_url: str = "https://www.wikidata.org/"):
        super().__init__(support_bbox=True, support_limit=True, support_offset=True, support_convert=True)
        self.request_url = request_url
        self.wikidata_client = Client(base_url=request_url)

    def get_entity(self, qid: str) -> WikidataEntity:
        qid_e = EntityId(qid)
        return WikidataEntity(qid_e, self.wikidata_client)

    # NOT IMPLEMENTED:

    def get_huts_from_source(
        self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: dict
    ) -> list[WikidataHutSource]:
        osm_service = OsmService()
        osm_huts = osm_service.get_huts_from_source(bbox=bbox, limit=limit * 10, offset=offset, **kwargs)
        huts = []
        for oh in osm_huts:
            tags = oh.source_data.tags if oh.source_data else None
            if tags is None:
                continue
            qid = tags.wikidata
            if not qid:
                continue
            logger.info(f" Wikidata entry {qid:<15} ({oh.name})")
            wikidata = wikidata_service.get_entity(qid)
            lon, lat = oh.location.lon_lat if oh.location else (None, None)
            wikidata_hut = WikidataHutSchema(
                id=qid,
                name=oh.name,
                lat=lat,
                lon=lon,
                attributes=wikidata.get_attributes(),
                photo=wikidata.get_photo(),
            )
            hut = WikidataHutSource(
                name=wikidata_hut.get_name(),
                source_id=wikidata_hut.get_id(),
                location=wikidata_hut.get_location(),
                source_data=wikidata_hut,
                source_properties=WikidataProperties(),
            )
            huts.append(hut)
            if len(huts) >= limit:
                break
        return huts

    def convert(self, src: t.Mapping | t.Any, include_photos: bool = True) -> HutSchema:
        hut_src = (
            WikidataHutSource(**src)
            if isinstance(src, t.Mapping)
            else WikidataHutSource.model_validate(src, from_attributes=True)
        )
        if hut_src.version >= 0:
            if hut_src.source_data is None:
                err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} without 'source_data' not allowed."
                raise AttributeError(err_msg)
            return WikidataHut0Convert(source_data=hut_src.source_data, include_photos=include_photos).get_hut()
        else:
            err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} not implemented."
            raise NotImplementedError(err_msg)


wikidata_service = WikidataService()

if __name__ == "__main__":
    logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)
    # qid = "Q42157530"
    # qid = "Q887175"  # Bluemlisalp
    # service = WikidataService()
    # entity = service.get_entity(qid)
    # rprint(entity.get_photos())

    limit = 10
    include_photos = True
    service = WikidataService()
    # service.get_wikidata_photos = wikidata_photos
    huts = service.get_huts_from_source(limit=limit)
    # licenses = {}
    # licenses_num: dict[str, int] = {}
    for h in huts:
        # rprint(h)
        hut = service.convert(h.model_dump(by_alias=True), include_photos=include_photos)
        rprint(hut)
        # if hut.photos:
        #     rprint(f"[bold]{hut.name.i18n}[/bold]")
        #     photo = hut.photos[0]
        #     # rprint(hut.extras.get("wikidata"))
        #     # filename = h.source_data.photo.title.replace("File:", "")
        #     # photo = get_wikicommon_image_info(filename)
        #     rprint(photo)
        #     rprint("--------------------------------")
        #     for lic in photo.licenses:
        #         licenses.update({lic.slug: lic.url})
        #         licenses_num[lic.slug] = licenses_num.get(lic.slug, 0) + 1
    # licenses_num = dict(sorted(licenses_num.items(), key=lambda x: x[1], reverse=True))
    # rprint(licenses)
    # rprint(licenses_num)



from typing import Annotated

from pydantic import BaseModel  # , Field
from pydantic.functional_validators import BeforeValidator


def none_to_str(v: str | None) -> str:
    """Convert a `None` to an empty string."""
    if v is None:
        return ""
    return v


NoneStr = Annotated[str, BeforeValidator(none_to_str)]


class TranslationSchema(BaseModel):
    """Field with different translations.

    Attributes:
        de: German
        en: English
        fr: French
    """

    de: NoneStr = ""
    en: NoneStr = ""
    fr: NoneStr = ""
    it: NoneStr = ""

    @property
    def i18n(self) -> str:
        """Returns the first stored translation in the following order:
        `de`, `en`, `fr` or `it`.
        """
        _codes: tuple = ("de", "en", "fr", "it")
        for code in _codes:
            if getattr(self, code, None):
                return str(getattr(self, code))
        return ""



from hut_services.core.schema import HutSchema
from hut_services.refuges_info import RefugesInfoService
from hut_services.refuges_info.schema import RefugesInfoHutSource


def test_refuges_info_service_source_online() -> None:
    """Get huts from opemstreemap.org"""
    limit = 2
    osm_service = RefugesInfoService()
    huts = osm_service.get_huts_from_source(limit=limit)
    assert len(huts) == 2
    for h in huts:
        # rprint(h)
        print(h.show(source_name=False))
        assert type(h) is RefugesInfoHutSource


def test_refuges_info_service_hut_online() -> None:
    """Tests conversion to HutSchema as well."""
    limit = 2
    osm_service = RefugesInfoService()
    huts = osm_service.get_huts(limit=limit)
    assert len(huts) == 2
    assert type(huts[0]) is HutSchema



import typing as t

from geojson_pydantic import Feature, FeatureCollection, Point  # noqa: F401
from geojson_pydantic.types import BBox, Position  # noqa: F401
from pydantic import BaseModel, ConfigDict, model_validator

from hut_services.core.utils import GPSConverter

from .types import Elevation, Latitude, Longitude


class LocationSchema(BaseModel):
    """Location with longitude, latitude and optional elevation in WSG84.

    Attributes:
        lon: Longitude (x).
        lat: Latitude (y).
        ele: Elevation in meter.
    """

    model_config = ConfigDict(from_attributes=True)

    lat: Latitude
    lon: Longitude
    # ele: Elevation | None = None

    @model_validator(mode="before")
    @classmethod
    def check_field_aliases(cls, data: t.Any) -> t.Any:
        for field, alias in [("lat", "y"), ("lon", "x"), ("ele", "z")]:
            if isinstance(data, dict) and alias in data and field not in data:
                data[field] = data[alias]
            elif hasattr(data, alias) and not hasattr(data, field):
                setattr(data, field, getattr(data, alias))
        return data

    @classmethod
    def from_swiss(cls, ch_lat: float, ch_lon: float, ele: float | None) -> "LocationSchema | LocationEleSchema":
        """Takes LV03 or LV95 (newest) coordiates.
        [More information](https://en.wikipedia.org/wiki/Swiss_coordinate_system).

        Args:
            ch_lat: Latitude in swiss format (east).
            ch_lon: Longitude in swiss format (north).
            ele: Elevation in meter.

        Returns:
            `LocationSchema` object.
        """
        if ele is None:
            ele = 0
        if ch_lat > 2000000:
            ch_lat -= 2000000
        if ch_lon > 1000000:
            ch_lon -= 1000000
        converter = GPSConverter()
        ch_lat, ch_lon, ele = converter.LV03toWGS84(ch_lat, ch_lon, ele)
        return cls(lat=ch_lat, lon=ch_lon)

    @property
    def lon_lat(self) -> tuple[Longitude, Latitude]:
        """Tuple with (`longitude`, `latitude`).

        Returns:
            Longitude and latitude.
        """
        return (self.lon, self.lat)

    @property
    def geojson(self) -> Point:
        """Retuns as geojson point."""
        return Point(coordinates=self.lon_lat, type="Point")

    def __str__(self) -> str:
        return f"lon={self.lon},lat={self.lat}"


class LocationEleSchema(LocationSchema):
    """Location with longitude, latitude and (optional) elevation in WSG84.

    Attributes:
        lat: Latitude.
        lon: Longitude.
        ele: Elevation in meter.
    """

    model_config = ConfigDict(from_attributes=True)

    ele: Elevation | None = None

    @classmethod
    def from_swiss(cls, ch_lat: float, ch_lon: float, ele: float | None) -> "LocationEleSchema":
        loc = super().from_swiss(ch_lat=ch_lat, ch_lon=ch_lon, ele=ele)
        return LocationEleSchema(lat=loc.lat, lon=loc.lon, ele=ele)

    @property
    def lon_lat_ele(self) -> tuple[Longitude, Latitude, Elevation]:
        """Tuple with longitude, latitude and elevation. If elevation is not defined it returns 0.

        Returns:
            Longitude, latitude and elevation."""
        ele = self.ele if self.ele else 0
        return (self.lon, self.lat, ele)

    @property
    def geojson(self) -> Point:
        """Retuns as geojson point."""
        return Point(coordinates=self.lon_lat_ele, type="Point")

    def __str__(self) -> str:
        return f"lon={self.lon},lat={self.lat},ele={self.ele}"



from collections.abc import Mapping, Sequence
from typing import Any

from pydantic import Field, model_validator
from slugify import slugify

from ._base import BaseSchema
from ._contact import ContactSchema
from ._hut_fields import CapacitySchema, HutTypeSchema, OpenMonthlySchema, OwnerSchema
from ._license import AuthorSchema, LicenseSchema, SourceSchema
from ._photo import PhotoSchema
from .geo import LocationEleSchema
from .locale import TranslationSchema

# import logging
# logger = logging.getLogger(__name__)


class HutSchema(BaseSchema):
    """Hut schema.

    Attributes:
        slug: Hut slug, if empty it is replaced by slugified `name`.
        name: Original hut name.
        location: Location of the hut.
        description: Description.
        notes: Additional public notes to the hut.
        owner: Hut owner.
        url: Hut website.
        contacts: Hut contacts.
        country_code: Country.
        comment: Additional private comment to the hut.
        capacity: Cpacities of the hut.
        type: Hut type (e.g. `selfhut`).
        photos: Hut photos.
        open_monthly: Monthly value if open, closed or partially open.
        is_active: Hut is active.
        is_public: Show hut public.
        extras: Additional information to the hut as dictionary.
        source: Short name of source (e.g. 'hrs' for alpsonline.org, or 'sac', 'refuges', ...).
    """

    slug: str = Field("", max_length=50)
    name: TranslationSchema = Field(..., description="Original hut name.")
    location: LocationEleSchema = Field(..., description="Location of the hut with optional elevation.")

    description: TranslationSchema = Field(default_factory=TranslationSchema)
    # description_attribution: str = Field("", max_length=1000, description="Description attribution/license.")
    license: LicenseSchema | None = Field(
        None,
        description="License for the hut information (mainly descpription)",
    )
    source: SourceSchema | None = Field(
        None, description="Source origin information (e.g. 'hrs' for alpsonline.org, or 'sac', 'refuges', ...)."
    )
    author: AuthorSchema | None = Field(
        None,
        description="Author of the hut information, this is an addition to the source information (e.g. custom author)",
    )

    notes: Sequence[TranslationSchema] = Field(..., description="Additional notes to the hut.")
    owner: OwnerSchema | None = Field(None)
    url: str = Field("", max_length=200)
    contacts: Sequence[ContactSchema] = Field(default_factory=list)
    country_code: str | None = Field(None, max_length=2, min_length=2)
    comment: str = Field("", max_length=20000, description="Additional private comment, e.g for review.")
    capacity: CapacitySchema
    hut_type: HutTypeSchema = Field(..., alias="type")
    photos: list[PhotoSchema] = Field(default_factory=list)
    open_monthly: OpenMonthlySchema
    is_active: bool = Field(default=True)
    is_public: bool = Field(default=True)
    extras: Mapping[str, Any] = Field(default_factory=dict, description="Additional information as dictionary.")
    # source: str = Field(
    #    "", max_length=20, description="Short name of source (e.g. 'hrs' for alpsonline.org, or 'sac', 'refuges', ...)."
    # )

    # infrastructure:       dict = Field(default_factory=dict, sa_column=Column(JSON)) # TODO, better name. Maybe use infra and service separated, external table
    # access:             Access = Field(default_factory=Access, sa_column=Access.get_sa_column())

    def __str__(self) -> str:
        slug = self.slug if self.slug else self.name.i18n.lower().replace(" ", "-")[:8]
        return f"<{slug} {self.name.i18n} ({self.location.lon},{self.location.lat})>"

    @model_validator(mode="after")
    def add_slug(self) -> "HutSchema":
        if not self.slug:
            self.slug = slugify(self.name.i18n, max_length=50, word_boundary=True)
        return self



# some coordiantes are not correct and are hardcoded here since the coordinates are used to uniquely identify a hut.
CORRECTIONS: dict[int, tuple[float, float]] = {
    # id: (lat     , lon    )   # hut name
    936745007: (45.95379, 7.15546),  # Cabane de la Tsissette
}



from hut_services import ContactSchema


def test_detect_phone_numbers_list() -> None:
    mobile_in = "079 522 3665"
    mobile_exp = "+41 79 522 36 65"
    phone_in = "062 7263232"
    phone_exp = "+41 62 726 32 32"
    phone, mobile = ContactSchema.number_to_phone_or_mobile([mobile_in, phone_in], region="CH")
    assert phone, phone_exp
    assert mobile, mobile_exp


def test_detect_phone_numbers_str() -> None:
    mobile_in = "079 522 3665"
    mobile_exp = "+41 79 522 36 65"
    phone_in = "062 7263232"
    phone_exp = "+41 62 726 32 32"
    phone, mobile = ContactSchema.number_to_phone_or_mobile(f"some text {mobile_in} or call {phone_in}", region="CH")
    assert phone, phone_exp
    assert mobile, mobile_exp



# some coordiantes are not correct and are hardcoded here since the coordinates are used to uniquely identify a hut.
CORRECTIONS: dict[int, tuple[float, float]] = {
    # id: (lat     , lon    )   # hut name
    2414: (46.52605, 8.17309),  # Oberaarjochhuette
    2980: (46.38300, 7.46782),  # Wildstrubelhuette
    3418: (46.48304, 7.12020),  # La Cabane des Choucas
    403: (46.30683, 6.74356),  # Refuge d'Ubine
    3605: (46.47469, 6.91070),  # le-Grillet
    419: (46.33056, 6.76587),  # Refuge de Bise
    2265: (45.98630, 7.63479),  # Lonza biwak
    5170: (46.2174989, 8.4489458),  # Bivacco-Campolatte-Baitin-dul-Peurat
    2336: (45.93388, 7.70813),  # Rifugio Testa Grigia Guide del Cervino
    2463: (46.15540, 7.27787),  # Cabane de Balavaux
    239: (45.89486, 6.98097),  # Refuge de Leschaux
    2382: (46.13942, 7.98927),  # Hohsaashütte
    2464: (46.07422, 7.86581),  # Mischabeljoch biwak
    3638: (46.05932, 7.06211),  # Cabane de Bovinette
}



class OSMCoordinatesError(Exception):
    def __init__(self, osm_id: int | str, name: str = ""):
        message = f"OSM coordinates missing for id {osm_id} (name: '{name}')"
        super().__init__(message)



__all__ = [
    "SERVICES",
    "AnswerEnum",
    "AuthorSchema",
    "BaseHutConverterSchema",
    "BaseHutSourceSchema",
    "BaseService",
    "CapacitySchema",
    "ContactSchema",
    "GeocodeService",
    "HutSchema",
    "HutSourceSchema",
    "HutTypeEnum",
    "HutTypeSchema",
    "LicenseSchema",
    "LocationEleSchema",
    "LocationSchema",
    "OpenMonthlySchema",
    "OsmService",
    "OwnerSchema",
    "PhotoSchema",
    "PhotoSchemaOld",
    "RefugesInfoService",
    "SourceDataSchema",
    "SourcePropertiesSchema",
    "SourceSchema",
    "TranslationSchema",
    "clear_file_cache",
    "file_cache",
]

from httpx import Auth

from .core.cache import clear_file_cache, file_cache
from .core.schema import (
    AnswerEnum,
    AuthorSchema,
    BaseHutConverterSchema,
    BaseHutSourceSchema,
    CapacitySchema,
    ContactSchema,
    HutSchema,
    HutSourceSchema,
    HutTypeEnum,
    HutTypeSchema,
    LicenseSchema,
    OpenMonthlySchema,
    OwnerSchema,
    PhotoSchema,
    PhotoSchemaOld,
    SourceDataSchema,
    SourcePropertiesSchema,
    SourceSchema,
)
from .core.schema.geo import LocationEleSchema, LocationSchema
from .core.schema.locale import TranslationSchema
from .core.service import BaseService
from .geocode import GeocodeService
from .osm import OsmService
from .refuges_info import RefugesInfoService
from .services import SERVICES



import pytest

from hut_services.core.schema.geo import LocationSchema
from hut_services.geocode.service import GeocodeService, geocode_service


def test_geocode_get_location() -> None:
    service = GeocodeService()
    name = "Almagellerhuette"
    coord = service.get_location_by_name(name)
    assert coord, "Nothing returned"
    assert pytest.approx(coord.lat) == 46.1076
    assert pytest.approx(coord.lon) == 8.007716


def test_geocode_get_elevations() -> None:
    service = geocode_service
    location = LocationSchema(lat=47.0, lon=11.1)
    location2 = LocationSchema(lat=46.0, lon=10.1)
    locations = [location, location2]
    eles = service.get_elevations(locations)
    assert eles, "Nothing returned"
    assert len(eles) == 2
    assert pytest.approx(eles[0].ele) == 2947
    assert pytest.approx(eles[1].ele) == 1087
    ele = service.get_elevation(location)
    assert pytest.approx(ele.ele) == eles[0].ele



#!/usr/bin/python2

#  The MIT License (MIT)
#
#  Copyright (c) 2014 Federal Office of Topography swisstopo, Wabern, CH and Aaron Schmocker
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#

# WGS84 <-> LV03 converter based on the scripts of swisstopo written for python2.7
# Aaron Schmocker [aaron@duckpond.ch]
# vim: tabstop=4 shiftwidth=4 softtabstop=4 expandtab

# Source: http://www.swisstopo.admin.ch/internet/swisstopo/en/home/topics/survey/sys/refsys/projections.html (see PDFs under "Documentation")
# Updated 9 dec 2014
# Please validate your results with NAVREF on-line service: http://www.swisstopo.admin.ch/internet/swisstopo/en/home/apps/calc/navref.html (difference ~ 1-2m)

# Downloaded from:
#
#   https://github.com/ValentinMinder/Swisstopo-WGS84-LV03/blob/master/scripts/py/wgs84_ch1903.py
import math

from hut_services.core.schema.geo.types import LV03, LV03X, LV03Y, Elevation, Latitude, Longitude


class GPSConverter:
    """
    GPS Converter class which is able to perform convertions between the
    CH1903 and WGS84 system.
    """

    # Convert CH y/x/h to WGS height
    def CHtoWGSheight(self, y: LV03Y, x: LV03X, h: Elevation) -> Elevation:
        # Axiliary values (% Bern)
        y_aux = (y - 600000) / 1000000
        x_aux = (x - 200000) / 1000000
        h = (h + 49.55) - (12.60 * y_aux) - (22.64 * x_aux)
        return h

    # Convert CH y/x to WGS lat
    def CHtoWGSlat(self, y: LV03Y, x: LV03X) -> Latitude:
        # Axiliary values (% Bern)
        y_aux = (y - 600000) / 1000000
        x_aux = (x - 200000) / 1000000
        lat = (
            (16.9023892 + (3.238272 * x_aux))
            + -(0.270978 * pow(y_aux, 2))
            + -(0.002528 * pow(x_aux, 2))
            + -(0.0447 * pow(y_aux, 2) * x_aux)
            + -(0.0140 * pow(x_aux, 3))
        )
        # Unit 10000" to 1" and convert seconds to degrees (dec)
        lat = (lat * 100) / 36
        return lat

    # Convert CH y/x to WGS long
    def CHtoWGSlng(self, y: LV03Y, x: LV03X) -> Latitude:
        # Axiliary values (% Bern)
        y_aux = (y - 600000) / 1000000
        x_aux = (x - 200000) / 1000000
        lng = (2.6779094 + (4.728982 * y_aux) + +(0.791484 * y_aux * x_aux) + +(0.1306 * y_aux * pow(x_aux, 2))) + -(
            0.0436 * pow(y_aux, 3)
        )
        # Unit 10000" to 1" and convert seconds to degrees (dec)
        lng = (lng * 100) / 36
        return lng

    # Convert decimal angle (° dec) to sexagesimal angle (dd.mmss,ss)
    def DecToSexAngle(self, dec: float) -> float:
        degree = int(math.floor(dec))
        minute = int(math.floor((dec - degree) * 60))
        second = (((dec - degree) * 60) - minute) * 60
        return degree + (float(minute) / 100) + (second / 10000)

    # Convert sexagesimal angle (dd.mmss,ss) to seconds
    def SexAngleToSeconds(self, dms: float) -> float:
        degree = 0.0
        minute = 0.0
        second = 0.0
        degree = math.floor(dms)
        minute = math.floor((dms - degree) * 100)
        second = (((dms - degree) * 100) - minute) * 100
        return second + (minute * 60) + (degree * 3600)

    # Convert sexagesimal angle (dd.mmss) to decimal angle (degrees)
    def SexToDecAngle(self, dms: float) -> float:
        degree = 0.0
        minute = 0.0
        second = 0.0
        degree = math.floor(dms)
        minute = math.floor((dms - degree) * 100)
        second = (((dms - degree) * 100) - minute) * 100
        return degree + (minute / 60) + (second / 3600)

    # Convert WGS lat/long (° dec) and height to CH h
    def WGStoCHh(self, lat: Latitude, lng: Longitude, ele: Elevation) -> Elevation:
        lat = self.DecToSexAngle(lat)
        lng = self.DecToSexAngle(lng)
        lat = self.SexAngleToSeconds(lat)
        lng = self.SexAngleToSeconds(lng)
        # Axiliary values (% Bern)
        lat_aux = (lat - 169028.66) / 10000
        lng_aux = (lng - 26782.5) / 10000
        ele = (ele - 49.55) + (2.73 * lng_aux) + (6.94 * lat_aux)
        return ele

    # Convert WGS lat/long (° dec) to CH x
    def WGStoCHx(self, lat: Latitude, lng: Longitude) -> LV03X:
        lat = self.DecToSexAngle(lat)
        lng = self.DecToSexAngle(lng)
        lat = self.SexAngleToSeconds(lat)
        lng = self.SexAngleToSeconds(lng)
        # Axiliary values (% Bern)
        lat_aux = (lat - 169028.66) / 10000
        lng_aux = (lng - 26782.5) / 10000
        x = (
            (200147.07 + (308807.95 * lat_aux) + +(3745.25 * pow(lng_aux, 2)) + +(76.63 * pow(lat_aux, 2)))
            + -(194.56 * pow(lng_aux, 2) * lat_aux)
        ) + +(119.79 * pow(lat_aux, 3))
        return x

    # Convert WGS lat/long (° dec) to CH y
    def WGStoCHy(self, lat: Latitude, lng: Longitude) -> LV03Y:
        lat = self.DecToSexAngle(lat)
        lng = self.DecToSexAngle(lng)
        lat = self.SexAngleToSeconds(lat)
        lng = self.SexAngleToSeconds(lng)
        # Axiliary values (% Bern)
        lat_aux = (lat - 169028.66) / 10000
        lng_aux = (lng - 26782.5) / 10000
        y = (
            (600072.37 + (211455.93 * lng_aux))
            + -(10938.51 * lng_aux * lat_aux)
            + -(0.36 * lng_aux * pow(lat_aux, 2))
            + -(44.54 * pow(lng_aux, 3))
        )
        return y

    def LV03toWGS84(self, east: float, north: float, height: Elevation) -> tuple[Latitude, Longitude, Elevation]:
        """
        Convert LV03 to WGS84 Return a array of double that contain lat, long,
        and height
        """
        return (self.CHtoWGSlat(east, north), self.CHtoWGSlng(east, north), self.CHtoWGSheight(east, north, height))

    def WGS84toLV03(self, latitude: Latitude, longitude: Longitude, ellHeight: Elevation) -> LV03:
        """
        Convert WGS84 to LV03 Return an array of double that contaign east,
        north, and height
        """
        return (
            self.WGStoCHy(latitude, longitude),
            self.WGStoCHx(latitude, longitude),
            self.WGStoCHh(latitude, longitude, ellHeight),
        )


if __name__ == "__main__":
    """Example usage for the GPSConverter class."""

    converter = GPSConverter()

    # Coordinates
    wgs84 = [46.95108, 7.438637, 0]
    # lv03: LV03 = ()

    # Convert WGS84 to LV03 coordinates
    lv03 = converter.WGS84toLV03(wgs84[0], wgs84[1], wgs84[2])

    print("WGS84: ")
    print(wgs84)
    print("LV03: ")
    print(lv03)
    # assert lv03 == (600000.3294154685, 199999.70729839153, -49.554241824400016)



import logging
from enum import Enum
from typing import Literal, cast

from geojson_pydantic import Feature, FeatureCollection, Point
from pydantic import BaseModel, Field, computed_field

from hut_services import (
    AuthorSchema,
    BaseHutConverterSchema,
    BaseHutSourceSchema,
    CapacitySchema,
    HutTypeEnum,
    HutTypeSchema,
    LicenseSchema,
    OwnerSchema,
    SourceDataSchema,
    SourcePropertiesSchema,
    SourceSchema,
)
from hut_services.core.guess import guess_hut_type
from hut_services.core.schema._photo import PhotoSchema
from hut_services.core.schema.geo import LocationEleSchema
from hut_services.core.schema.locale import TranslationSchema

from .coordinates import CORRECTIONS
from .utils import get_original_images, refuges_lic

logger = logging.getLogger(__name__)


class HutTypeRefugesEnum(str, Enum):
    missing = "missing"
    cabane_non_gardee = "cabane-non-gardee"
    refuge_garde = "refuge-garde"
    gite_d_etape = "gite-d-etape"
    batiment_en_montagne = "batiment-en-montagne"


REFUGES_HUT_TYPES: dict[int, HutTypeRefugesEnum] = {
    7: HutTypeRefugesEnum.cabane_non_gardee,
    10: HutTypeRefugesEnum.refuge_garde,
    9: HutTypeRefugesEnum.gite_d_etape,
    28: HutTypeRefugesEnum.batiment_en_montagne,
}
WODORE_HUT_TYPES: dict[int, HutTypeEnum] = {
    7: HutTypeEnum.selfhut,
    10: HutTypeEnum.hut,
    9: HutTypeEnum.hut,
    28: HutTypeEnum.bhotel,
}


class RefugesInfoProperties(SourcePropertiesSchema):
    """Properties save together with the source data."""

    slug: str = Field(..., description="hut slug by refuges.info")
    hut_type: HutTypeRefugesEnum = Field(..., description="hut type by refuges.info.")


class _Coord(BaseModel):
    alt: float
    long: float
    lat: float
    precision: dict[str, str]


class _ValeurNom(BaseModel):
    nom: str
    valeur: str | None


class _ValeurID(BaseModel):
    ident: int = Field(..., alias="id")
    valeur: str | None


class _NomID(BaseModel):
    ident: int = Field(..., alias="id")
    nom: str | None


class _Type(_ValeurID):
    icone: str


class _Etat(_ValeurID):
    ident: Literal["ouverture", "fermeture", "cle_a_recuperer", "detruit", "NULL", "null"] | None = Field(
        ..., alias="id"
    )  # type: ignore[assignment]


class _Date(BaseModel):
    derniere_modif: str | None
    creation: str


class _Article(BaseModel):
    demonstratif: str
    defini: str
    partitif: str


class _SiteOfficiel(_ValeurNom):
    url: str | None


class _PlacesMatelas(_ValeurNom):
    nb: int | None


class _InfoComp(BaseModel):
    site_officiel: _SiteOfficiel
    manque_un_mur: _ValeurNom
    cheminee: _ValeurNom
    poele: _ValeurNom
    couvertures: _ValeurNom
    places_matelas: _PlacesMatelas
    latrines: _ValeurNom
    bois: _ValeurNom
    eau: _ValeurNom


class _Description(BaseModel):
    valeur: str


class _RefugesInfoFeatureProperties(BaseModel):
    ident: int = Field(..., alias="id")
    lien: str  # link
    nom: str
    sym: str
    coord: _Coord
    hut_type: _Type = Field(..., alias="type")
    places: _ValeurNom
    etat: _Etat
    date: _Date
    remarque: _ValeurNom
    acces: _ValeurNom
    proprio: _ValeurNom
    createur: _NomID
    article: _Article
    info_comp: _InfoComp
    description: _Description


class RefugesInfoFeature(Feature, SourceDataSchema):
    """RefugesInfo Feature Model with required properties and geometry."""

    geometry: Point
    properties: _RefugesInfoFeatureProperties

    def get_id(self) -> str:
        return str(self.properties.ident)

    def get_name(self) -> str:
        return str(self.properties.nom).strip().strip('"').strip()

    def get_location(self) -> LocationEleSchema:
        # coords = self.geometry.coordinates
        lat: float | None
        lon: float | None
        lat, lon = CORRECTIONS.get(self.properties.ident, (self.properties.coord.lat, self.properties.coord.long))
        return LocationEleSchema(lat=lat, lon=lon, ele=self.properties.coord.alt)

    def get_properties(self) -> RefugesInfoProperties:
        slug = self.properties.lien.split("/")[-2]
        return RefugesInfoProperties(
            hut_type=REFUGES_HUT_TYPES.get(self.properties.hut_type.ident, HutTypeRefugesEnum.missing), slug=slug
        )


class RefugesInfoFeatureCollection(FeatureCollection):
    """Used to get FeatureCollection, not returned by the service."""

    generator: str
    copyright_by: str = Field(..., alias="copyright")
    timestamp: str
    size: str
    features: list[RefugesInfoFeature]


class RefugesInfoHutSource(BaseHutSourceSchema[RefugesInfoFeature, RefugesInfoProperties]):
    """Data from refuges.info database."""

    source_name: str = "refuges"


class RefugesInfoHut0Convert(BaseHutConverterSchema[RefugesInfoFeature]):
    @property
    def _props(self) -> _RefugesInfoFeatureProperties:
        return self.source_data.properties

    ## implemented in base
    # @computed_field  # type: ignore[prop-decorator]
    # @property
    # def slug(self) -> str:
    #    return f"refuges-{self.source.get_id()}"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def name(self) -> TranslationSchema:
        return TranslationSchema(fr=self.source_data.get_name(), de=self.source_data.get_name())

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source_name(self) -> str:
        return "refuges"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def description(self) -> TranslationSchema:
        # return TranslationSchema(fr=self._props.description.valeur or "")
        return TranslationSchema(fr=self._props.remarque.valeur or "")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def author(self) -> AuthorSchema | None:
        if self.description.fr:
            return AuthorSchema(name="Les refuges.info contributors", url="https://www.refuges.info")
        else:
            return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source(self) -> SourceSchema | None:
        return SourceSchema(
            name=self.source_name,
            ident=self.source_data.get_id(),
            url=self._props.lien,
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def license(self) -> LicenseSchema | None:
        return refuges_lic

    @computed_field  # type: ignore[prop-decorator]
    @property
    def notes(self) -> list[TranslationSchema]:
        _note_fr = self._props.remarque.valeur or ""
        _note_de = ""
        if self._props.etat == "cle_a_recuperer":
            _note_fr = f"Clés nécessaires \n\n{_note_fr}".strip()
            _note_de = "Schlüssel erforderlich"
        if _note_fr:
            return [TranslationSchema(fr=_note_fr, de=_note_de)]
        return []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def owner(self) -> OwnerSchema | None:
        return None

    @computed_field()  # type: ignore[prop-decorator]
    @property
    def photos(self) -> list[PhotoSchema]:
        if self.include_photos is False:
            return []
        return cast(list[PhotoSchema], get_original_images(self.source_data.get_id()))

    @computed_field  # type: ignore[prop-decorator]
    @property
    def url(self) -> str:
        return self._props.info_comp.site_officiel.url or ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def capacity(self) -> CapacitySchema:
        try:
            return CapacitySchema(open=int(self._props.places.valeur), closed=None)  # type: ignore  # noqa: PGH003
        except TypeError:
            return CapacitySchema(open=None, closed=None)

    @computed_field(alias="type")  # type: ignore[prop-decorator]
    @property
    def hut_type(self) -> HutTypeSchema:
        return guess_hut_type(
            name=self.name.i18n or "",
            default=WODORE_HUT_TYPES.get(self._props.hut_type.ident, HutTypeEnum.unknown),
            capacity=self.capacity,
            elevation=self.location.ele,
            operator=None,
            missing_walls=self._props.info_comp.manque_un_mur.valeur or "0",
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_public(self) -> bool:
        return self._props.etat.ident in ["ouverture", "cle_a_recuperer"] or self._props.etat.ident is None



import logging
from typing import Any

from pydantic import Field, computed_field

from hut_services import (
    BaseHutConverterSchema,
    BaseHutSourceSchema,
    LicenseSchema,
    LocationEleSchema,
    PhotoSchema,
    SourceDataSchema,
    SourcePropertiesSchema,
    SourceSchema,
    TranslationSchema,
)
from hut_services.core.schema import BaseSchema
from hut_services.core.schema.geo.types import Latitude, Longitude
from hut_services.wikicommons.service import wikicommons_service

logger = logging.getLogger(__name__)


class WikidataPhotoInfo(BaseSchema):
    size: int
    width: int
    height: int
    url: str
    descriptionurl: str
    descriptionshorturl: str
    mime: str


class WikidataPhotoAttributes(BaseSchema):
    ns: int
    title: str
    missing: str
    known: str
    imagerepository: str
    imageinfo: list[WikidataPhotoInfo]
    contentmodel: str
    pagelanguage: str
    pagelanguagehtmlcode: str
    pagelanguagedir: str
    fullurl: str
    editurl: str
    canonicalurl: str


class WikidataPhoto(BaseSchema):
    """Wikidata photo."""

    title: str
    attributes: WikidataPhotoAttributes


class WikidataHutSchema(SourceDataSchema):
    """Open street map schema."""

    attributes: dict[str, Any]
    photo: WikidataPhoto | None
    wikidata_id: str = Field(..., alias="id")
    name: str
    lat: Latitude | None
    lon: Longitude | None

    def get_id(self) -> str:
        """Get open street map `id`."""
        return str(self.wikidata_id)

    def get_name(self) -> str:
        """Get open street map hut name."""
        return self.name

    def get_location(self) -> LocationEleSchema:
        """Get open street map location."""
        if self.lat is not None and self.lon is not None:
            return LocationEleSchema(lat=self.lat, lon=self.lon)
        else:
            return LocationEleSchema(lat=0, lon=0)


class WikidataProperties(SourcePropertiesSchema):
    pass


class WikidataHutSource(BaseHutSourceSchema[WikidataHutSchema, WikidataProperties]):
    version: int = 0
    source_name: str = "wikidata"


class WikidataHut0Convert(BaseHutConverterSchema[WikidataHutSchema]):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def name(self) -> TranslationSchema:
        return TranslationSchema(de=self.source_data.get_name())

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source_name(self) -> str:
        return "wikidata"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def description(self) -> TranslationSchema:
        return TranslationSchema()

    @computed_field()  # type: ignore[prop-decorator]
    @property
    def photos(self) -> list[PhotoSchema]:
        image = self.source_data.photo
        if image is None or self.include_photos is False:
            return []
        return [wikicommons_service.get_photo(image.title.replace("File:", ""))]

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source(self) -> SourceSchema | None:
        return SourceSchema(
            name=self.source_name,
            ident=self.source_data.get_id(),
            url=f"https://www.wikidata.org/wiki/{self.source_data.get_id()}",
        )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def license(self) -> LicenseSchema | None:
        return LicenseSchema(
            slug="cc-by-sa-4.0", name="CC-BY-SA 4.0", url="https://creativecommons.org/licenses/by-sa/4.0/"
        )



from pathlib import Path

import toml

from tasks import Ctx, EnvError, echo, env, error, header, info, success, task, warning  # noqa: F401


@task
def update_venv(c: Ctx, dry: bool = False):
    """Updated venv activate script with custom commands"""
    commands = [
        "# custom command",
        'alias t="inv -p"',
        'alias tl="inv --list"',
        "source <(inv --print-completion-script bash)",
        "complete -F _complete_invoke -o default invoke inv t",
    ]
    for cmd in commands:
        info(f"Append '{cmd}' to activate script")
        c.run(f"echo '{cmd}' >> .venv/bin/activate") if not dry else None
    success("Updated venv activate script") if not dry else success("Would have updated venv activate script")
    info("Run 'source .venv/bin/activate'")


@task(help={"venv_update": "Updates venv activate script (runs per default)"})
def install(c: Ctx, venv_update: bool = True):
    """Install the uv environment and install the pre-commit hooks"""
    echo("🚀 Creating virtual environment using pyenv and poetry")
    c.run("uv sync")
    c.run("uv run pre-commit install")
    if venv_update:
        update_venv(c)
    success("Installation done, your are ready to go ...")


@task(
    help={
        "add_tag": "Add the new tag and push it to remote",
        "unreleased": "Show only unreleased changes",
        "length": "Number of lines to show (mutual exclusive with unreleased)",
    }
)
def release(
    c: Ctx,
    add_tag: bool = False,
    dry: bool = False,
    unreleased: bool = True,
    length: int = -1,
):
    """Prepare a release, update CHANGELOG file and bump versions"""
    new_tag = c.run("git-cliff --bumped-version", hide=True).stdout.strip()
    new_version = new_tag.replace("v", "")
    if dry:
        unreleased = unreleased if length < 0 else False
        length = 0 if (not unreleased and length < 0) else length
        (
            echo("Show only 'unreleased' changes")
            if unreleased
            else (echo(f"Show {length} lines of changelog") if length > 0 else echo("Show full changelog"))
        )
        header("Changelog start") if dry else None
        cl = (
            c.run(
                f"git-cliff --bump {'--unreleased' if unreleased else ''}",
                hide=True,
            )
            .stdout.strip()
            .split("\n")
        )
        if unreleased or length == 0:
            echo("\n".join(cl))
        else:
            line = "\n".join(cl[:length])
            echo(f"{line}\n...\n")
        header("Changelog end") if dry else None
    else:
        cl = c.run("git-cliff --bump -u --prepend CHANGELOG.md", hide=True).stdout.strip().split("\n")
        c.run(f"bump2version --new-version {new_version} patch")

        # only prepend new tag -- this way it is possible to edit it.
        # cl = c.run(f"git-cliff --bump {'--prepend CHANGELOG.md' if dry else '-o'}", hide=True).stdout.strip().split("\n")
    if new_tag:
        success(f"Bumped to version '{new_version}' (tag '{new_tag}').")
        info("Please check the entries in 'CHANGELOG.md' and update it accordingly.")
        if add_tag and not dry:
            c.run(f"git tag -f '{new_tag}'")
            c.run(f"git push origin '{new_tag}'")
    else:
        warning("Did not update to new version tag.")


@task(help={"next": "Show next version"})
def version(c: Ctx, next: bool = False):  # noqa: A002
    """Print current or next (--next) project version"""
    version = "unknown"
    # adopt path to your pyproject.toml
    pyproject_toml_file = Path(__file__).parent.parent / "pyproject.toml"
    if pyproject_toml_file.exists() and pyproject_toml_file.is_file():
        data = toml.load(pyproject_toml_file)
        # check project.version
        if "project" in data and "version" in data["project"]:
            version = data["project"]["version"]
        # check tool.poetry.version
        elif "tool" in data and "poetry" in data["tool"] and "version" in data["tool"]["poetry"]:
            version = data["tool"]["poetry"]["version"]
    if next:
        new_version = c.run("git-cliff --bumped-version", hide=True).stdout.strip().replace("v", "")
        if new_version == version:
            warning("No newer version available")
        echo(new_version)
        return
    echo(version)



from hut_services import BaseService, HutSourceSchema

from .osm.service import OsmService
from .refuges_info.service import RefugesInfoService
from .wikidata.service import WikidataService

SERVICES: dict[str, BaseService[HutSourceSchema]] = {
    "osm": OsmService(),
    "refuges": RefugesInfoService(),
    "wikidata": WikidataService(),
}



from ._base import BaseSchema
from ._booking import BookingSchema, HutBookingsSchema, OccupancyStatusEnum, PlacesSchema, ReservationStatusEnum
from ._contact import ContactSchema
from ._hut import HutSchema
from ._hut_base_converter import BaseHutConverterSchema
from ._hut_base_source import BaseHutSourceSchema, HutSourceSchema, SourceDataSchema, SourcePropertiesSchema
from ._hut_fields import (
    AnswerEnum,
    CapacitySchema,
    HutTypeEnum,
    HutTypeSchema,
    OpenMonthlySchema,
    OwnerSchema,
    PhotoSchemaOld,
)
from ._license import AuthorSchema, LicenseSchema, SourceSchema
from ._photo import PhotoSchema

__all__ = [
    "AnswerEnum",
    "AuthorSchema",
    "BaseHutConverterSchema",
    "BaseHutSourceSchema",
    "BookingSchema",
    "CapacitySchema",
    "ContactSchema",
    "HutBookingsSchema",
    "HutSchema",
    "HutTypeEnum",
    "HutTypeSchema",
    "LicenseSchema",
    "OccupancyStatusEnum",
    "OpenMonthlySchema",
    "OwnerSchema",
    "PhotoSchema",
    "PhotoSchemaOld",
    "PlacesSchema",
    "ReservationStatusEnum",
    "SourceDataSchema",
    "SourcePropertiesSchema",
    "SourceSchema",
]



#!/usr/bin/env python
# from functools import lru_cache
import logging
import typing as t

import httpx

# from typing import Any, Literal, Mapping
from hut_services import BaseService, file_cache
from hut_services.core.schema import HutSchema
from hut_services.core.schema.geo.geo import LocationEleSchema, LocationSchema
from hut_services.geocode.schema import GeocodeHut0Convert, GeocodeHutSchema, GeocodeHutSource

# from hut_services.core.schema.geo import BBox

if __name__ == "__main__":  # only for testing
    from icecream import ic  # type: ignore[import-untyped] # noqa: F401, RUF100 , PGH003
    from rich import print as rprint  # noqa: F401, RUF100
    from rich.traceback import install

    install(show_locals=False)


logger = logging.getLogger(__name__)


@file_cache(ignore=["client"], expire_in_seconds=60 * 24 * 7 * 4)  # one month
def _get_location_by_name(
    name: str,
    request_url: str,
    client: httpx.Client | None = None,
    countries: t.Sequence[str] = ["ch", "de", "fr", "it", "at"],
    languages: t.Sequence[str] | None = None,
) -> t.Any:
    logger.debug(f"Get location for {name} from '{request_url}'.")
    if client is None:
        client = httpx.Client()
    country_codes = ",".join(countries)
    accept_lang = ",".join(languages) if languages else country_codes
    params: dict[str, str | int | bool] = {
        "q": name,
        "format": "jsonv2",
        "limit": 1,
        "extratags": True,
        "countrycodes": country_codes,
        "accept-language": accept_lang,
    }
    return client.get(url=request_url, params=params).json()


@file_cache(ignore=["client"], expire_in_seconds=60 * 24 * 7 * 4 * 12)  # 12 months
def _get_elevations(
    locations: t.Sequence[LocationSchema | LocationEleSchema],
    request_url: str,
    client: httpx.Client | None = None,
) -> t.Any:
    logger.debug(f"Get elevation for {locations} from '{request_url}'.")
    if client is None:
        client = httpx.Client()
    params: dict[str, str | int | bool] = {
        "locations": "|".join([f"{location.lat},{location.lon}" for location in locations])
    }
    # Post does somehow not work
    # data = {"locations": [{"latitude": loc.lat, "longitude": loc.lon} for loc in locations]}
    # res = client.post(url=request_url, data=data, headers=headers)
    headers = {"Content-Type": "application/json"}
    res = client.get(url=request_url, params=params, headers=headers)
    return res.json().get("results", [])


class GeocodeService(BaseService[GeocodeHutSource]):
    """Service to get Information from
    [Nominatim](https://nominatim.openstreetmap.org).

    Note:
        This is not used to get huts, rather to get additional info (like location) from huts.
    """

    def __init__(self) -> None:
        super().__init__(support_bbox=True, support_limit=True, support_offset=True, support_convert=True)
        self.loc_request_url = "https://nominatim.openstreetmap.org/search"
        self.ele_request_url = "https://api.open-elevation.com/api/v1/lookup"
        self.httpx_client = httpx.Client()

    def get_location_by_name(self, name: str, client: httpx.Client | None = None) -> LocationEleSchema | None:
        """Get location (coordinates) from a name (uses 'https://nominatim.openstreetmap.org)."""
        client = client if client is not None else self.httpx_client
        res = _get_location_by_name(name=name, request_url=self.loc_request_url, client=client)
        if res:
            return GeocodeHutSchema(**res[0]).get_location()
        return None

    def get_elevations(
        self, locations: t.Sequence[LocationSchema | LocationEleSchema], client: httpx.Client | None = None
    ) -> list[LocationEleSchema]:
        """Get elevations from a list of locations (coordinates) (uses 'https://open-elevation.com)."""
        client = client if client is not None else self.httpx_client
        res = _get_elevations(locations=locations, request_url=self.ele_request_url, client=client)
        locs: list[LocationEleSchema] = []
        for i, loc in enumerate(locations):
            ele = res[i].get("elevation")
            locs.append(LocationEleSchema(lat=loc.lat, lon=loc.lon, ele=ele))
        return locs

    def get_elevation(
        self, location: LocationSchema | LocationEleSchema, client: httpx.Client | None = None
    ) -> LocationEleSchema:
        """Get elevation from a location (coordinates) (uses 'https://open-elevation.com)."""
        client = client if client is not None else self.httpx_client
        return self.get_elevations(locations=[location], client=client)[0]

    # NOT IMPLEMENTED:
    # def get_huts_from_source(
    #    self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: dict
    # ) -> list[GeocodeHutSource]:

    def convert(self, src: t.Mapping | t.Any, include_photos: bool = True) -> HutSchema:
        hut_src = (
            GeocodeHutSource(**src)
            if isinstance(src, t.Mapping)
            else GeocodeHutSource.model_validate(src, from_attributes=True)
        )
        if hut_src.version >= 0:
            if hut_src.source_data is None:
                err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} without 'source_data' not allowed."
                raise AttributeError(err_msg)
            return GeocodeHut0Convert(source_data=hut_src.source_data, include_photos=include_photos).get_hut()
        else:
            err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} not implemented."
            raise NotImplementedError(err_msg)


geocode_service = GeocodeService()

if __name__ == "__main__":
    logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
    # qid = "Q42157530"
    # qid = "Q887175"  # Bluemlisalp
    # service = WikidataService()
    # entity = service.get_entity(qid)
    # rprint(entity.get_photos())
    limit = 5
    service = GeocodeService()
    loc = service.get_location_by_name("Allmagelerhuette")
    rprint(loc)
    if loc:
        eles = service.get_elevations([loc])
        rprint(eles)

    ## service.get_wikidata_photos = wikidata_photos
    # huts = service.get_huts_from_source(limit=limit)
    # for h in huts:
    #    # rprint(h)
    #    hut = service.convert(h.model_dump(by_alias=True))
    #    rprint(hut.name.i18n)
    #    # rprint(hut.extras.get("wikidata"))
    #    rprint(hut.photos)



#!/usr/bin/env python
# from functools import lru_cache
import logging
import textwrap
import typing as t

import overpy  # type: ignore[import-untyped]

from hut_services.core.cache import file_cache
from hut_services.core.schema import HutSchema
from hut_services.core.schema.geo import BBox
from hut_services.core.service import BaseService
from hut_services.osm.schema import OsmHut0Convert, OsmHutSchema, OsmHutSource, OsmProperties

if __name__ == "__main__":  # only for testing
    from rich import print as rprint  # noqa: F401, RUF100
    from rich.traceback import install

    install(show_locals=False)


logger = logging.getLogger(__name__)


@file_cache(ignore=["api"])
def _get_huts_from_source(
    api: t.Any, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: dict
) -> list[OsmHutSource]:
    if bbox is None:
        # fetch all ways and nodes
        # SWISS
        lon = [45.7553, 47.6203]
        lat = [5.7127, 10.5796]
        bounder = 100.0
        lon_diff = lon[1] - lon[0]
        lat_diff = lat[1] - lat[0]
        lon_range = lon_diff / bounder * limit + lon_diff / bounder * 2 * offset
        lon_range = lon_diff if lon_range > lon_diff else lon_range
        lat_range = lat_diff / bounder * limit + lat_diff / bounder * 2 * offset
        lat_range = lat_diff if lat_range > lat_diff else lat_range
        lon_start = lon[0] + (lon_diff - lon_range) / 2
        lat_start = lat[0] + (lat_diff - lat_range) / 2
        bbox = (lon_start, lat_start, lon_start + lon_range, lat_start + lat_range)
    area = ",".join([str(b) for b in bbox])  # f"{lon_start},{lat_start},{lon_start+lon_range},{lat_start+lat_range}"
    logger.info(f"get osm data from {api.url} with bbox: ({area})")
    query = f"""
            [out:json];
            (
            nw["tourism"="alpine_hut"]["name"]({area});
            nw["tourism"="wilderness_hut"]["name"]({area});
            );
            out qt center {limit};
        """
    logger.debug(f"query:\n{'-' * 20}\n{textwrap.dedent(query).strip()}\n{'-' * 20}")
    try:
        result = api.query(query)
    except overpy.exception.OverpassGatewayTimeout as e:
        logger.warning("overpy execution failed")
        logger.debug(str(e))
        return []
    huts = []
    for _osm_type, res in {"node": result.nodes, "way": result.ways}.items():
        osm_type: t.Literal["node", "way", "area"] = _osm_type  # type: ignore  # noqa: PGH003 # ignore pyright and mypy str to literal assigment
        for h in res:
            osm_hut = OsmHutSchema.model_validate(h, from_attributes=True)
            osm_hut.osm_type = osm_type
            hut = OsmHutSource(
                name=osm_hut.get_name(),
                source_id=osm_hut.get_id(),
                location=osm_hut.get_location(),
                source_data=osm_hut,
                source_properties=OsmProperties(osm_type=osm_type),
            )
            huts.append(hut)
            # huts.append(h)
    logger.info(f"succesfully got {len(huts)} huts")
    return huts


class OsmService(BaseService[OsmHutSource]):
    """Service to get huts from
    [Open Street Map](https://www.openstreetmap.org/)
    with [overpass api](https://wiki.openstreetmap.org/wiki/Overpass_API).

    Note:
        The methods are descriebed in [`BaseService`][hut_services.BaseService].
    """

    def __init__(self, request_url: str = "https://overpass.osm.ch/api/"):
        super().__init__(support_bbox=True, support_limit=True, support_offset=True, support_convert=True)
        self.request_url = request_url

    def get_huts_from_source(
        self, bbox: BBox | None = None, limit: int = 1, offset: int = 0, **kwargs: dict
    ) -> list[OsmHutSource]:
        api = overpy.Overpass(url=self.request_url)
        huts = _get_huts_from_source(api=api, bbox=bbox, limit=limit, offset=offset, **kwargs)
        assert all(isinstance(p, OsmHutSource) for p in huts), "Wrong type, not a list of 'PhotoSchema'"  # noqa: S101
        return t.cast(list[OsmHutSource], huts)

    def convert(self, src: t.Mapping | t.Any, include_photos: bool = True) -> HutSchema:
        hut_src = (
            OsmHutSource(**src)
            if isinstance(src, t.Mapping)
            else OsmHutSource.model_validate(src, from_attributes=True)
        )
        if hut_src.version >= 0:
            if hut_src.source_data is None:
                err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} without 'source_data' not allowed."
                raise AttributeError(err_msg)
            return OsmHut0Convert(include_photos=include_photos, source_data=hut_src.source_data).get_hut()
        else:
            err_msg = f"Conversion for '{hut_src.source_name}' version {hut_src.version} not implemented."
            raise NotImplementedError(err_msg)


if __name__ == "__main__":
    logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)
    limit = 10
    wikidata_photos = True
    osm_service = OsmService()
    # osm_service.get_wikidata_photos = wikidata_photos
    huts = osm_service.get_huts_from_source(limit=limit)
    for h in huts:
        # rprint(h)
        hut = osm_service.convert(h.model_dump(by_alias=True))
        rprint("====================")
        rprint(hut.name.i18n)
        if hut.photos:
            rprint(hut.photos)
        # rprint(h.source_properties_schema)
        # rprint(h)
        # print(h.show(source_name=False))



from pydantic import BaseModel, ConfigDict


class BaseSchema(BaseModel):
    """Use this as a base for all schemas, sets `from_attributes` to `True`."""

    model_config = ConfigDict(from_attributes=True, extra="allow")



# from hut_services.core.utils import GPSConverter
#
#
# def test_wgs84_to_lv03_swiss() -> None:
#    converter = GPSConverter()
#
#    # Coordinates
#    wgs84 = [46.95108, 7.438637, 0]
#    # Convert WGS84 to LV03 coordinates
#    lv03 = converter.WGS84toLV03(wgs84[0], wgs84[1], wgs84[2])
#
#    print("WGS84: ")
#    print(wgs84)
#    print("LV03: ")
#    print(lv03)
#    assert lv03 == (600000.3294154685, 199999.70729839153, -49.554241824400016)



from rich import print as rprint

from hut_services import OsmService

osm_service = OsmService()
osm_service.get_wikidata_photos = True  # include photos
huts = osm_service.get_huts_from_source(limit=5)  # get the first 10 entries
for h in huts:
    hut = osm_service.convert(h.model_dump(by_alias=True))
    rprint(hut.name.i18n)
    rprint(hut.url) if hut.url else None
    if hut.photos:
        rprint(hut.photos)



# type: ignore  # noqa: PGH003
import httpx
from rich import print as rprint

request_url = (
    "https://nominatim.openstreetmap.org/search"  # ?q=Neue%20Regensburger%20H%C3%BCtte&polygon_geojson=0&format=jsonv2"
)

langs = ",".join(["ch", "de", "fr", "it", "at"])
params = {
    "amenity": "Neue Regensburger Huette",
    "format": "jsonv2",
    "limit": 1,
    "extratags": True,
    "countrycodes": langs,
    "accept-language": langs,
}

res = httpx.get(url=request_url, params=params).json()
if res:
    hut = res[0]
rprint(hut)
